/*
<LICENCE>

Copyright (c) 2007, University of Southampton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this
	list of conditions and the following disclaimer.

  *	Redistributions in binary form must reproduce the above copyright notice,
	this list of conditions and the following disclaimer in the documentation
	and/or other materials provided with the distribution.

  *	Neither the name of the University of Southampton nor the names of its
	contributors may be used to endorse or promote products derived from this
	software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

</LICENCE>
*/

package r2q2.content;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Iterator;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import javax.activation.DataHandler;
import javax.activation.MimetypesFileTypeMap;
import javax.xml.soap.AttachmentPart;
import javax.xml.soap.SOAPException;

import org.apache.axis.Message;
import org.apache.axis.MessageContext;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.XMLOutputter;
import org.xml.sax.InputSource;

import com.gavinwillingham.util.properties.NoSuchPropertyException;
import com.gavinwillingham.util.properties.PropertiesLoader;

import r2q2.NoAttachmentException;
import r2q2.util.streams.ByteArrayDataSource;
import r2q2.util.streams.StreamCopier;
import r2q2.util.xml.DocumentBuilder;

/**
 * Web service to unpack content packages
 *
 * @author Gavin Willingham
 */
public class ContentUnpacker {
	
	private EncodedFileMap files = new EncodedFileMap();
	private String httpRoot;

    /**
     * Unpacks a QTI content package
     * 
     * <p>
     * Takes a content package as a SOAP attachment and returns a list of
     * filenames (including directory names), and a set of SOAP attachments
     * which are the files themselves (not including directories)
     * </p>
     * 
     * TODO: 	Test if file is *ZIP* file, both check the MIME and fallover 
     * 			handling
     * TODO: Javadoc Exceptions
     *
     * @param _xmlName filename of the Item XML
     * @param _httpRoot the http-referenceable base path for the files
     * @return file and directory names
     * @throws NoAttachmentException 
     * @throws UnpackingException 
     */
    public EncodedFileList unpack(String _xmlName, String _httpRoot) throws NoAttachmentException, UnpackingException {
        if(_xmlName == null || _xmlName.length() == 0){
        	throw new UnpackingException("XML name not provided");
        }
        if(_httpRoot == null || _httpRoot.length() == 0){
        	//TODO: test valid URL
        	throw new UnpackingException("HTTP root not provided");
        }
    	
    	if(!_httpRoot.endsWith("/"))
        	this.httpRoot = "/" + _httpRoot;
        else
        	this.httpRoot = _httpRoot;
    	
    	//Get the request and response objects, and the attachment (zip file)
        MessageContext msgContext = MessageContext.getCurrentContext();
        Message request = msgContext.getRequestMessage();
        Message response = msgContext.getResponseMessage();
        Iterator attachmentIter = request.getAttachments();

        //test if a zip was included
        if (!attachmentIter.hasNext()) {
           throw new NoAttachmentException("Zip file not attached");
        } else {
            AttachmentPart zip = (AttachmentPart)attachmentIter.next();
            try {
                //Build zip processing objects
                DataHandler dh = zip.getDataHandler();
                ZipInputStream in = new ZipInputStream(dh.getInputStream());
                ZipEntry entry;
                
                //loop over entries in zip file
                while ((entry = in.getNextEntry()) != null) {
                    //add the file/directory name to the file list
                    String filename = entry.getName();
                    byte[] file = StreamCopier.copyToByteArray(in);
                    try {
						files.put(new ByteFile(file, filename));
					} catch (DuplicateFileException e) {
						/* If this line is reached, someone's sending some 
						 * weird ZIP files.. this shouldn't be possible
						 */ 
						throw new UnpackingException("Zip file entry '" + filename + "' is a duplicate!",e);
					}
                }
                
                //tidy up the zip stream
                in.close();
                
                //ensure all files exist
                collateFiles(_xmlName);
                
                return generateResponse(response, _xmlName);
                
            } catch (IOException e) {
				e.printStackTrace();
				throw new UnpackingException(e);
			} catch (SOAPException e) {
				e.printStackTrace();
				throw new UnpackingException(e);
			} catch (JDOMException e) {
				e.printStackTrace();
				throw new UnpackingException(e);
			} catch (UnavailableFileException e) {
				e.printStackTrace();
				throw new UnpackingException(e);
			} catch (UnsupportedProtocolException e) {
				e.printStackTrace();
				throw new UnpackingException(e);
			} catch (NoSuchFileException e) {
				e.printStackTrace();
				throw new UnpackingException(e);
			} catch (NoSuchPropertyException e) {
				e.printStackTrace();
				throw new UnpackingException(e);
			}
        }
        
    }
    
    /**
     * TODO: javadoc
     * TODO: write this method
     * 
     * @return
     * @throws NoSuchFileException 
     * @throws DuplicateFileException 
     * @throws NoSuchPropertyException 
     * @throws UnsupportedProtocolException 
     * @throws UnavailableFileException 
     * @throws IOException 
     * @throws JDOMException 
     */
    public EncodedFileList collateFiles(String _xml, String _xmlName, String _httpRoot) throws NoSuchFileException, DuplicateFileException, JDOMException, IOException, UnavailableFileException, UnsupportedProtocolException, NoSuchPropertyException{
    	
    	//add xml to FileMap
    	//TODO encoding issues
    	files.put(new ByteFile(_xml.getBytes(), _xmlName));
    	
    	//ensure all files exist
        collateFiles(_xmlName);
        
        //Get the response object
        MessageContext msgContext = MessageContext.getCurrentContext();
        Message response = msgContext.getResponseMessage();
        
        return generateResponse(response, _xmlName);
    }
    
    private EncodedFileList generateResponse(Message response, String _xmlName) throws NoSuchFileException {
        //Generate list of filenames and attachments to return
        MimetypesFileTypeMap mimes = new MimetypesFileTypeMap();
        Iterator<ByteFile> iter = files.iterator();
        while(iter.hasNext()){
        	ByteFile file = iter.next();
        	String filename = file.name;
        	if(!(filename.endsWith("/") || filename.endsWith("\\"))){ 
        		//file, not directory
        		ByteArrayDataSource bads = 
                    new ByteArrayDataSource(filename, file.bytes, 
                                            mimes.getContentType(filename));
                AttachmentPart responseAttachment = response.createAttachmentPart();
                responseAttachment.setDataHandler(new DataHandler(bads));
                response.addAttachmentPart(responseAttachment);
        	}
        }
        
        return files.getFileList(_xmlName);
	}

	/**
     * TODO: javadoc
     * TODO: exception handling
     * 
     * @param _xml
     * @return
     * @throws JDOMException
     * @throws IOException
     * @throws UnsupportedProtocolException 
     * @throws UnavailableFileException 
     * @throws NoSuchPropertyException 
     */
    private void collateFiles(String _xmlName) throws JDOMException, IOException, UnavailableFileException, UnsupportedProtocolException, NoSuchPropertyException{
    	
    	//get the XML file
        byte[] xml;
        try{
        	xml = files.getBytes(_xmlName);
        }
        catch(NoSuchFileException e){
        	throw new UnavailableFileException("Base XML file (" + _xmlName + ") not present", e);
        }
    	Document doc = DocumentBuilder.getDocument(xml);
        
        //process xml
        processElement(doc.getRootElement());
    	
        //convert new XML to byte[]
        XMLOutputter out = new XMLOutputter();
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        out.output(doc, baos);
        ByteFile bf = new ByteFile(baos.toByteArray(), _xmlName);
        try {
			files.put(bf, true);
		} catch (DuplicateFileException e1) {
			/* TODO: does this need a proper exception? Theoretically 
			 * impossible to reach
			 */
			e1.printStackTrace();
		}
    }
    
    /**
     * TODO: javadoc
     * 
     * @param _e
     * @param _files
     * @throws UnsupportedProtocolException 
     * @throws UnavailableFileException 
     */
    private void processElement(Element _e) throws UnavailableFileException, UnsupportedProtocolException{
    	//process children
    	Iterator children = _e.getChildren().iterator();
    	while(children.hasNext()){
    		processElement((Element)children.next());
    	}
    	
    	//process self
    	String elementName = _e.getName();
    	if(elementName.equals("img")){
    		updateAttributeURI(_e, "src");
    	}
    	else if(elementName.equals("responseProcessing")){
    		/*TODO: should we be checking for unsupported templates here? or
    		 * 		inserting URLs to URI-based template references
    		 */
    		updateAttributeURI(_e, "templateLocation");
    	}
    	else if(elementName.equals("object")){
    		updateAttributeURI(_e, "data");
    	}
    }
    
    /**
     * TODO: javadoc
     * 
     * @param _e
     * @param _attrib
     * @param _files
     * @throws UnsupportedProtocolException 
     * @throws UnavailableFileException 
     */
    private void updateAttributeURI(Element _e, String _attrib) throws UnavailableFileException, UnsupportedProtocolException{
    	String uri = _e.getAttributeValue(_attrib);
		if(uri != null){
			//get encoded name, downloading if required
    		String newName = getEncodedFileName(uri);
    		
    		//update XML
    		_e.setAttribute(_attrib, httpRoot + newName);		    		
		}    	
    }  
    
    /**
     * TODO: javadoc
     * @param _uri
     * @param _files
     * @return
     * @throws UnavailableFileException 
     * @throws UnsupportedProtocolException 
     */   
    private String getEncodedFileName(String _uri) throws UnavailableFileException, UnsupportedProtocolException{
    	String out_ = null;
    	
    	try{
	    	if(!_uri.matches("[a-zA-Z]+:\\/\\/.*")){
	    		//relative path
	    		out_ = files.getEncodedName(_uri);
	    	}
	    	else if(_uri.startsWith("http://")){
	    		//if remote (HTTP) file
	    		
	    		//download the file
	    		ByteFile bf = getRemoteFile(_uri);
	    		
	    		/* It is possible that file has the same name as existing file, 
	    		 * but is unique. This loops until a unique name has been 
	    		 * obtained in the format filename[i] where i is an incremented
	    		 * integer
	    		 */
	    		String name = bf.name;
	    		int i=0;
	    		while(true){
		    		try{
		    			out_ = files.put(bf);
		    			break;
		    		} catch (DuplicateFileException e) {
		    			bf.name = name + "[" + i++ + "]";
		    		}
	    		}
	    	}
	    	else if(_uri.startsWith("file://")){
	    		//if local file
	    		/* TODO: 	very odd thing to do... we shouldn't be supporting 
	    		 * 			full paths for security reasons, but relative ones 
	    		 * 			possibly. Here we just strip off the file:// and 
	    		 * 			look for the remainder... should probably consider 
	    		 * 			this in more depth... of course, I may just be 
	    		 * 			going crazy - who knows?
	    		 */
	    		out_ = files.getEncodedName(_uri.substring(7));
	    	}
	    	else {
	    		//Some unknown protocol
	    		throw new UnsupportedProtocolException("Unsupported protocol in URI: " + _uri);
	    	}
    	} catch (NoSuchFileException e) {
			e.printStackTrace();
			throw new UnavailableFileException("Local file does not exist", e);
		}
		
    	return out_;
    }
    
    /**
     * Retrieves a remote file via HTTP
     * 
     * @param _uri URI of remote file to download
     * @return The bytes and (path-less) filename of the requested file
     * @throws UnavailableFileException if file cannot be obtained
     */
    private ByteFile getRemoteFile(String _uri) throws UnavailableFileException{
    	byte[] bytes;
    	
    	try {
			URL uri = new URL(_uri);
			InputStream is = uri.openStream();
			bytes = StreamCopier.copyToByteArray(is);
		} catch (MalformedURLException e) {
			throw new UnavailableFileException("Malformed URL in remote file location", e);
		} catch (IOException e) {
			throw new UnavailableFileException("IO Exception when reading remote file", e);
		}
		
		//generate the local name
		//get the filename (strip off the path)
		String localName = _uri.replaceFirst("^.*[(\\/|\\\\]", "");
		
		//if an image, stick in the images directory
		//TODO: better organisation, other file types
		if(localName.endsWith(".png")
			|| localName.endsWith(".jpg")
			|| localName.endsWith(".jpeg")
			|| localName.endsWith(".gif"))
			localName = "images/" + localName;
		
    	return new ByteFile(bytes, localName);
    }  
}
