/*
<LICENCE>

Copyright (c) 2007, University of Southampton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this
	list of conditions and the following disclaimer.

  *	Redistributions in binary form must reproduce the above copyright notice,
	this list of conditions and the following disclaimer in the documentation
	and/or other materials provided with the distribution.

  *	Neither the name of the University of Southampton nor the names of its
	contributors may be used to endorse or promote products derived from this
	software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

</LICENCE>
*/

package r2q2.content;

import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Random;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.twmacinta.util.MD5;

/**
 * Wraps a HashMap for mapping filenames to encoded versions and file bytes. 
 * 
 * Also stores a base path from which all contained (relative) paths descend. Caller is
 * responsible for ensuring that base path is unique.
 * 
 * TODO: Javadoc individual methods
 * 
 * @author Gavin Willingham
 */
class EncodedFileMap{

	private String basePath;
	private LinkedHashMap<String,ByteFile> filenames;
	
	private static Random rand = new Random(System.currentTimeMillis());
	private static Pattern filePattern = Pattern.compile("(.*(\\/|\\\\))*([^(\\/|\\\\)]*$)");
	private static Pattern dirPattern = Pattern.compile("(.*(\\/|\\\\))*([^(\\/|\\\\)]*(\\/|\\\\)$)");

	/**
	 * Default constructor, no base path set
	 *
	 */
	public EncodedFileMap(){
		filenames = new LinkedHashMap<String,ByteFile>();
	}
	
	/**
	 * Constructor that sets the base path
	 * @param _basePath the base path
	 */
	public EncodedFileMap(String _basePath){ 
		this();
		this.basePath = _basePath; 
	}
	
	public String getBasePath() {
		return basePath;
	}
	public void setBasePath(String basePath) {
		this.basePath = basePath;
	}
	
	/**
	 * TODO: error handling on regex
	 * 
	 * @param _filename
	 * @return
	 */
	private String put(String _filename, byte[] _bytes){
		String encodedFilename;
		
    	duplicate: while(true){//loop on tiny chance that duplicate filename is generated
    		String prefix = "";
    		String suffix = "";
    		String md5String = null;
    		Matcher m;
    		
    		//generate random part
            MD5 md5 = new MD5();
        	md5.Update(String.valueOf(rand.nextDouble()));
        	md5String = md5.asHex();
        	
	    	//get the unencoded (relative) path to the file
			if(_filename.endsWith("\\") || _filename.endsWith("/")){
				suffix = "/";
				m = dirPattern.matcher(_filename);
			}
			else{
				String[] filenameParts = _filename.split("\\.");
				suffix = "." + filenameParts[filenameParts.length - 1];
				m = filePattern.matcher(_filename);
				
			}
			//get the encoded version
			if(!m.matches()){
				//TODO: throw proper exception
				return null;
			}
			String dir = m.group(1);
			if(dir != null){
				/* test that parent directories exist - important for 
				 * externally obtained files
				 */
				ByteFile dirFile = filenames.get(dir);
				if(dirFile == null){
					prefix = put(dir, null);
				}
				else{
					prefix = dirFile.name;
				}
			}
			
			//build the new path and filename
			encodedFilename = prefix + md5String + suffix;
			
			Iterator iter = filenames.values().iterator();
			while(iter.hasNext()){
				ByteFile out_ = (ByteFile)iter.next();
				if(out_.name.equals(encodedFilename)) continue duplicate;
			}
			
			break;
    	}
    	
    	ByteFile bf = new ByteFile(_bytes, encodedFilename);
		filenames.put(_filename, bf);
		
		return encodedFilename;
	}
	
	public String put(ByteFile _bf) throws DuplicateFileException{
		return this.put(_bf, false);
	}
	
	public String put(ByteFile _bf, boolean _overwrite) throws DuplicateFileException{
		String filename = _bf.name;
		if(filenames.get(filename) != null){
			if(_overwrite){
				filenames.remove(filename);
			}
			else{
				//throw an exception, filename already exists
				throw new DuplicateFileException(filename);
			}
		}
		return this.put(filename, _bf.bytes);
	}
	
	/**
	 * Gets the encoded name for the specified file
	 * @param _filename the unencoded file name
	 * @return the encoded file name
	 * @throws NoSuchFileException if the requested file is not present
	 */
	public String getEncodedName(String _filename) throws NoSuchFileException{
		ByteFile out_ = filenames.get(_filename);
		if(out_ == null) throw new NoSuchFileException(_filename);
		return out_.name;
	}
	
	/**
	 * Gets the file bytes for the specified (unencoded) filename
	 * 
	 * @param _filename the unencoded filename
	 * @return the file bytes
	 * @throws NoSuchFileException if the requested file is not present
	 */
	public byte[] getBytes(String _filename) throws NoSuchFileException{
		ByteFile out_ = this.filenames.get(_filename);
		if(out_ == null) throw new NoSuchFileException(_filename);
		return out_.bytes;
	}

	/**
	 * @return an Iterator containing the ByteFile objects
	 */
	public Iterator<ByteFile> iterator(){
		return filenames.values().iterator();
	}
	
	/**
	 * javadoc
	 * @param _xmlName The name of the XML file within the EncodedFileMap
	 * @return 	A new EncodedFileList object, containing the encoded names and 
	 * 			the XMl file name
	 * @throws NoSuchFileException 
	 */
	public EncodedFileList getFileList(String _xmlName) throws NoSuchFileException{
		ByteFile[] files = filenames.values().toArray(new ByteFile[0]);
		String[] encodedFileNames = new String[files.length];
		for(int i=0; i<files.length; i++) encodedFileNames[i] = files[i].name;
		String xmlName = this.getEncodedName(_xmlName);
		return new EncodedFileList(xmlName, encodedFileNames);
	}
}
