/*
<LICENCE>

Copyright (c) 2007, University of Southampton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this
	list of conditions and the following disclaimer.

  *	Redistributions in binary form must reproduce the above copyright notice,
	this list of conditions and the following disclaimer in the documentation
	and/or other materials provided with the distribution.

  *	Neither the name of the University of Southampton nor the names of its
	contributors may be used to endorse or promote products derived from this
	software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

</LICENCE>
*/

package r2q2.processing.expression.operators;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;

import org.jdom.Element;

import r2q2.processing.expression.Expression;
import r2q2.processing.expression.ExpressionFactory;
import r2q2.processing.expression.InvalidArgumentsException;
import r2q2.processing.util.OrderableVector;
import r2q2.variable.ItemVariable;
import r2q2.variable.ItemVariable.Cardinality;
import r2q2.variable.ItemVariable.BaseType;
import r2q2.variable.TransitoryVariable;

/**
 * 
 * Contains operator expression
 * 
 * constraints: 
 *   takes a two sub expressions of the same base type
 *   must have the same cardinality, single or multiple
 *   returns Boolean 'true' if the first expression is contained within
 *   the container of the second
 *   
 *   ordered containers behave differently from ordered containers:
 *		-ordered containers must have the order in the same sequence in both
 *		-unordered containers contain all elements in any order
 *   if sub expression is NULL, NULL is returned
 *   should NOT be used on base-type of float due to poor definition
 *      --taking the above as an exception worthy violation, design decision noted
 *   MUST NOT be used on sub-expressions of base-type duration
 */
public class Contains implements Expression{
	
	private Expression subpart[];
	
	public Contains(Element _e) throws InvalidArgumentsException
	{
		subpart = new Expression[2];
		//get from the factory an expression for each child element (Node)
		ExpressionFactory of = ExpressionFactory.getInstance();
		Iterator children = _e.getChildren().iterator();
    	int i=0;
		try{
			while(children.hasNext()){
	    		if(i < 2)
	    		{
					subpart[i] = of.makeExpression((Element)children.next());
	    		}
		    	i++;
	    	}
    	}catch(InvalidArgumentsException iae)
    	{
    		throw new InvalidArgumentsException("Error occured in building childrens expression: ",iae);
    	}
    	if( i < 2 || i > 2)
    		throw new InvalidArgumentsException("Error: usage violation, Member must be called with exactly 2 argument rather than the "+i+" arguments specified");
	}
	
	public ItemVariable eval(HashMap<String,ItemVariable> vars) throws InvalidArgumentsException
	{
		try{
			ItemVariable test1 = subpart[0].eval(vars);
			ItemVariable test2 = subpart[1].eval(vars);
			
			if(test1 == null || test2 == null)
				return null;
			//first the cardinality test
			if((test1.card == Cardinality.multiple || test1.card == Cardinality.ordered) && 
			   (test2.card == Cardinality.multiple || test2.card == Cardinality.ordered))
			{
				Vector t1Vec = (Vector)test1.value;
				Vector t2Vec = (Vector)test2.value;
				
				//tests on type occur here
				if(t1Vec.size() > 0 && t2Vec.size() > 0)
				{	
					if(test1.varType != test2.varType || test1.varType == BaseType.afloat || test1.varType == BaseType.duration)
						throw new InvalidArgumentsException("Error: arguements of differing or constraint violating type, aborting");		
				}
				
				if(test1.card == Cardinality.ordered && test2.card == Cardinality.ordered)
				{
					Boolean found = new Boolean(false);
					
					for(int i=0; i < t1Vec.size(); i++)
					{
						if(t1Vec.elementAt(i).equals(t2Vec.elementAt(0)))
						{
							int j = 1;
							while(j < t2Vec.size() && t1Vec.elementAt(i+j).equals(t2Vec.elementAt(j)))
								j++;
							if(j == t2Vec.size())
								found = new Boolean(true);
						}
					}
					return new TransitoryVariable(BaseType.bool,Cardinality.single,found);
				}
				else if(test1.card == Cardinality.multiple && test2.card == Cardinality.multiple)
				{
					Boolean found = new Boolean(true);
					
					int i=0;
					while(i < t2Vec.size() && found.booleanValue())
					{
						boolean fnd = false;
						int j =0;
						while (!fnd && j < t1Vec.size())
						{
							if(t2Vec.elementAt(i).equals(t1Vec.elementAt(j)))
							{
								t1Vec.removeElement(t2Vec.elementAt(i));
								fnd = true;
							}else
								j++;
						}
						if(!fnd)
							found = new Boolean(false);
						else
							i++;
					}
					return new TransitoryVariable(BaseType.bool,Cardinality.single,found);
				}
				else
					throw new InvalidArgumentsException("Error: arguements of differing ordering type, aborting");
			}
			else 
				throw new InvalidArgumentsException("Error: arguements are of differing or unsuitable cardinality, aborting");
		}
		catch(InvalidArgumentsException iae)
		{
			throw new InvalidArgumentsException("Error: evaluation of container yielded unexpected exception",iae);
		}
	}
}