/*
<LICENCE>

Copyright (c) 2007, University of Southampton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this
	list of conditions and the following disclaimer.

  *	Redistributions in binary form must reproduce the above copyright notice,
	this list of conditions and the following disclaimer in the documentation
	and/or other materials provided with the distribution.

  *	Neither the name of the University of Southampton nor the names of its
	contributors may be used to endorse or promote products derived from this
	software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

</LICENCE>
*/

package r2q2.processing.expression.operators;

import java.util.HashMap;
import java.util.Iterator;
import java.util.StringTokenizer;

import org.jdom.Element;

import r2q2.processing.expression.Expression;
import r2q2.processing.expression.ExpressionFactory;
import r2q2.processing.expression.InvalidArgumentsException;
import r2q2.variable.ItemVariable;
import r2q2.variable.ItemVariable.Cardinality;
import r2q2.variable.TransitoryVariable;
import r2q2.variable.ItemVariable.BaseType;
import r2q2.variable.ItemVariable.Cardinality;
/**
 * 
 * 'equalRounded' operator expression
 * 
 * constraints: 
 *   takes a two sub expressions of numeric base type
 *   expressions must have single cardinality
 *   returns Boolean 'true' if the two expressions are numerically equal
 *   after rounding
 *   if either sub expression is NULL, NULL is returned
 */

public class EqualRounded implements Expression{
	
	public enum RoundingMode {significantFigures, decimalPlaces}
	
	private Expression subpart[];
	private RoundingMode roundingMode;
	private int figures;
	
	public EqualRounded(Element _e) throws InvalidArgumentsException
	{
		subpart = new Expression[2];
		//get from the factory an expression for each child element (Node)
		ExpressionFactory of = ExpressionFactory.getInstance();
		Iterator children = _e.getChildren().iterator();
		
		String roundModString = _e.getAttributeValue("roundingMode");
		if(roundModString.equals("decimalPlaces"))
			roundingMode = RoundingMode.decimalPlaces;
		else if(roundModString.equals("significantFigures"))
			roundingMode = RoundingMode.significantFigures;
		else
			throw new InvalidArgumentsException("Error: unknown rounding mode attribute : "+roundModString+", aborting");
		
		if(_e.getAttributeValue("figures") != null)
			figures = Integer.parseInt(_e.getAttributeValue("figures"));		
		
    	int i=0;
		try{
			while(children.hasNext()){
	    		if(i < 2)
	    		{
					subpart[i] = of.makeExpression((Element)children.next());
	    		}
		    	i++;
	    	}
    	}catch(InvalidArgumentsException iae)
    	{
    		throw new InvalidArgumentsException("Error occured in building childrens expression: ",iae);
    	}
    	if( i < 2 || i > 2)
    		throw new InvalidArgumentsException("Error: usage violation, 'equalRounded' must be called with exactly 2 argument rather than the "+i+" arguments specified");
	}
	
	public ItemVariable eval(HashMap<String,ItemVariable> vars) throws InvalidArgumentsException
	{
		try{
			ItemVariable test1 = subpart[0].eval(vars);
			ItemVariable test2 = subpart[1].eval(vars);
			
			//simple escape clause for null terms
			if(test1 == null || test2 == null)
				return null;
			
			//cardinality check
			if(test1.card != Cardinality.single || test2.card != Cardinality.single)
				throw new InvalidArgumentsException("Error: the 'equal' operation can only be called on two operands each of single cardinality, aborting");
			
			//first attempt to convert to a numerical format
			if(test1.varType == BaseType.integer && test2.varType == BaseType.integer)
			{
				Integer i1 = (Integer)test1.value;
				Integer i2 = (Integer)test2.value;
				
				if(roundingMode == RoundingMode.decimalPlaces)
				{
					if(i1.intValue() == i2.intValue())
						return new TransitoryVariable(BaseType.bool,Cardinality.single, new Boolean(true));
					else
						return new TransitoryVariable(BaseType.bool,Cardinality.single, new Boolean(false));
				}else 
				{
					if( sigFig(i1,figures) == sigFig(i2,figures))
						return new TransitoryVariable(BaseType.bool,Cardinality.single, new Boolean(true));
					else
						return new TransitoryVariable(BaseType.bool,Cardinality.single, new Boolean(false));
				}
			}
			else
			{
				Double f1;
				Double f2;
				if(test1.varType == BaseType.integer)
					f1 = new Double(((Integer)test1.value).intValue());
				else if (test1.varType == BaseType.afloat)
					f1 = (Double)test1.value;
				else
					throw new InvalidArgumentsException("Error: first sub expression is not of a supported numeric type, aborting");
				
				if(test2.varType == BaseType.integer)
					f2 = new Double(((Integer)test2.value).intValue());
				else if (test2.varType == BaseType.afloat)
					f2 = (Double)test2.value;
				else
					throw new InvalidArgumentsException("Error: second sub expression is not of a supported numeric type, aborting");
				
				if(roundingMode == RoundingMode.decimalPlaces)
				{
					if(roundDouble(f1, figures) == roundDouble(f2,figures))
						return new TransitoryVariable(BaseType.bool,Cardinality.single, new Boolean(true));
					else
						return new TransitoryVariable(BaseType.bool,Cardinality.single, new Boolean(false));
				}else 
				{
					if( sigFig(f1,figures) == sigFig(f2,figures))
						return new TransitoryVariable(BaseType.bool,Cardinality.single, new Boolean(true));
					else
						return new TransitoryVariable(BaseType.bool,Cardinality.single, new Boolean(false));
				}
			}	
		}
		catch(InvalidArgumentsException iae)
		{
			throw new InvalidArgumentsException("Error: evaluation of sub expressions yielded unexpected exception",iae);
		}
	}
	
	//rounds to set amount of decimal places
	public static final double roundDouble(double d, int places) 
	{
		return Math.round(d * Math.pow(10, (double) places)) / Math.pow(10,(double) places);
	}
	
	
	//dirty way of abrieviating a number to 'sf' amount of significant figures
	private static double sigFig(double f, int sf)
	{
		String tmp = f+"";
		//count to the decimal place
		int len=0;
		while(len < tmp.length() && tmp.charAt(len) != '.')
			len++;
		
		String trunc = "";
		String exp = tmp.substring(tmp.indexOf("E"));
		
		for(int i=0; i<sf; i++)
		{
			if(tmp.charAt(i) == '.')
				sf+=1;
			trunc += tmp.charAt(i);
		}
		if(tmp.charAt(sf) > '4')
			trunc = trunc.substring(0,trunc.length()-1)+(((trunc.charAt(sf-1))+1)- 48);
		
		double ret = Double.parseDouble(trunc+exp);
		return ret;
	}

	//tests the sig fig calculation
	public static void main(String args[])
	{
		//double test = (double)12345678.125;
		//double after = sigFig(test,10);
		//System.out.println("after is : "+after);
		
		Double testD = new Double(6.8);
		System.out.println((int)Math.ceil(testD.doubleValue()));
		System.out.println((int)Math.round(testD.doubleValue()));
		Double testD2 = new Double(-6.8);
		System.out.println((int)Math.ceil(testD2.doubleValue()));
		System.out.println((int)Math.round(testD2.doubleValue()));
	}
}