/*
<LICENCE>

Copyright (c) 2007, University of Southampton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this
	list of conditions and the following disclaimer.

  *	Redistributions in binary form must reproduce the above copyright notice,
	this list of conditions and the following disclaimer in the documentation
	and/or other materials provided with the distribution.

  *	Neither the name of the University of Southampton nor the names of its
	contributors may be used to endorse or promote products derived from this
	software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

</LICENCE>
*/

package r2q2.processing.expression.operators;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;

import org.jdom.Element;

import r2q2.processing.expression.Expression;
import r2q2.processing.expression.ExpressionFactory;
import r2q2.processing.expression.InvalidArgumentsException;
import r2q2.processing.util.OrderableVector;
import r2q2.variable.ItemVariable;
import r2q2.variable.TransitoryVariable;
import r2q2.variable.ItemVariable.BaseType;
import r2q2.variable.ItemVariable.Cardinality;

/**
 * 
 * product operator expression
 * 
 * constraints: 
 *   takes one or more expressions of numerical base type
 *   sub expressions must have single cardinality
 *   returns a float product of all expressions OR an integer product if all expresions are of type Integer
 *   if any of the sub expressions are NULL, NULL is returned
 */
public class Product implements Expression{
	
	private Vector<Expression> subparts;
	
	public Product(Element _e) throws InvalidArgumentsException
	{
		subparts = new Vector<Expression>();
		//get from the factory an expression for each child element (Node)
		ExpressionFactory of = ExpressionFactory.getInstance();
		Iterator children = _e.getChildren().iterator();
		int i = 0;
		try{
			while(children.hasNext()){
				subparts.add(of.makeExpression((Element)children.next()));
		    	i++;
	    	}
    	}catch(InvalidArgumentsException iae)
    	{
    		throw new InvalidArgumentsException("Error occured in building childrens expression: ",iae);
    	}
    	if( i < 1)
    		throw new InvalidArgumentsException("Error: usage violation, 'product' must be called with at least 1 argument, aborting");
	}
	
	public ItemVariable eval(HashMap<String,ItemVariable> vars) throws InvalidArgumentsException
	{
		try{
			boolean floatSeen = false;
			Double retD = new Double(0);
			Integer retI = new Integer(0);
			for(int i=0; i<subparts.size(); i++)
			{
				ItemVariable test = subparts.elementAt(i).eval(vars);
				if(test == null)
					return null;
				
				if(test.card != Cardinality.single)
					throw new InvalidArgumentsException("Error: 'product' can only take single cardinality arguements, aborting");
				if(test.varType == BaseType.afloat)
				{
					if(!floatSeen)
					{
						floatSeen = true;
						retD = new Double(((Double)test.value).doubleValue()*retI.doubleValue());
					}
					else
						retD = new Double(((Double)test.value).doubleValue()*retD.doubleValue());
				}
				else if(test.varType == BaseType.integer)
				{
					if(!floatSeen)
						retI = new Integer(((Integer)test.value).intValue()*retI.intValue());
					else
						retD = new Double(((Integer)test.value).doubleValue()*retD.doubleValue());
				}
				else
					throw new InvalidArgumentsException("Error: 'product' can only take single cardinality numeric types, aborting");
			}
			if(!floatSeen)
				return new TransitoryVariable(BaseType.integer,Cardinality.single,retI);
			else
				return new TransitoryVariable(BaseType.afloat,Cardinality.single,retD);
		}
		catch(InvalidArgumentsException iae)
		{
			throw new InvalidArgumentsException("Error: evaluation of container yielded unexpected exception",iae);
		}
	}
}