/*
<LICENCE>

Copyright (c) 2007, University of Southampton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this
	list of conditions and the following disclaimer.

  *	Redistributions in binary form must reproduce the above copyright notice,
	this list of conditions and the following disclaimer in the documentation
	and/or other materials provided with the distribution.

  *	Neither the name of the University of Southampton nor the names of its
	contributors may be used to endorse or promote products derived from this
	software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

</LICENCE>
*/

package r2q2.rendering.xhtml;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import org.jdom.Attribute;
import org.jdom.Content;
import org.jdom.DataConversionException;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.Namespace;
import org.jdom.Text;

import com.gavinwillingham.util.properties.NoSuchPropertyException;

import r2q2.SpecificationViolationException;
import r2q2.initialisation.InitialisationEngine;
import r2q2.rendering.Renderer;
import r2q2.rendering.RendererResponse;
import r2q2.rendering.RenderingEngineResponse;
import r2q2.rendering.RenderingException;
import r2q2.rendering.xhtml.processors.AssociateInteractionProcessor;
import r2q2.rendering.xhtml.processors.ChoiceInteractionProcessor;
import r2q2.rendering.xhtml.processors.GapMatchInteractionProcessor;
import r2q2.rendering.xhtml.processors.GraphicAssociateInteractionProcessor;
import r2q2.rendering.xhtml.processors.GraphicGapMatchInteractionProcessor;
import r2q2.rendering.xhtml.processors.GraphicOrderInteractionProcessor;
import r2q2.rendering.xhtml.processors.HotspotInteractionProcessor;
import r2q2.rendering.xhtml.processors.HottextInteractionProcessor;
import r2q2.rendering.xhtml.processors.InlineChoiceInteractionProcessor;
import r2q2.rendering.xhtml.processors.InteractionProcessor;
import r2q2.rendering.xhtml.processors.MatchInteractionProcessor;
import r2q2.rendering.xhtml.processors.OrderInteractionProcessor;
import r2q2.rendering.xhtml.processors.PositionObjectInteractionProcessor;
import r2q2.rendering.xhtml.processors.SelectPointInteractionProcessor;
import r2q2.rendering.xhtml.processors.SliderInteractionProcessor;
import r2q2.rendering.xhtml.processors.TextEntryInteractionProcessor;
import r2q2.util.streams.StreamCopier;
import r2q2.util.xml.DocumentBuilder;

/**
 * TODO javadoc 
 * 
 * @author Gavin Willingham
 *
 */
public class XHTMLRenderer implements Renderer {
	
	//MIME type of the output byte array
	public static final String mimeType = "text/html";
	
	//Output strings
	private String htmlString = "";
	private String depth = "";
	
	//Variables & settings
	public HashMap[] vars;
	public HashMap<String, List<List<String>>> opts;
	public List<String> variableNames;
	public HashSet<String> inTag = new HashSet<String>();
	public boolean adaptive = false;
	public boolean firstStage = false;
	public InteractionProcessor currentIntProc = null;
	
	public static final Namespace ns = Namespace.getNamespace("http://www.imsglobal.org/xsd/imsqti_v2p0");
	
	//Tag sets
	private HashSet<String> interactionTags = null;
	private HashSet<String> feedbackTags = null;
	private HashSet<String> templateTags = null;
	private HashSet<String> xhtmlTags = null;
	private HashSet<String> attrBody = null;
	
	public XHTMLRenderer(){
		initialiseTagSets();
	}

	public RendererResponse render(byte[] _xml, HashMap[] _variables,
			HashMap<String, List<List<String>>> _options, List<String> _variableNames, boolean _firstStage) throws RenderingException, SpecificationViolationException {
		
		vars = _variables;
		opts = _options;
		variableNames = _variableNames;
		firstStage = _firstStage;
		
		if(opts == null) opts = new HashMap<String,List<List<String>>>();
		if(variableNames == null) variableNames = new ArrayList<String>();
		
		Element root = null;
		try {
			root = DocumentBuilder.getRootElement(_xml);
		} catch (JDOMException e) {
			throw new RenderingException(e);
		} catch (IOException e) {
			throw new RenderingException(e);
		} catch (NoSuchPropertyException e) {
			throw new RenderingException(e);
		}
		
		//Set the adaptive flag
		try {
			Attribute adap = root.getAttribute("adaptive");
			if(adap != null) adaptive = adap.getBooleanValue();
		} catch (DataConversionException e2) {
			throw new SpecificationViolationException("Adaptive attribute not a boolean");
		}
		
		//Set the prolog and html header stuff
		addToOutput("<?xml version=\"1.0\" encoding=\"UTF-8\"?>", false);
		addToOutput("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\"");
		addToOutput("\"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd \">", false);
		decreaseDepth();
		addToOutput("<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">");
		
		//build the header
		String title = root.getAttributeValue("title");
		addToOutput("<head>");
		addToOutput("<title>" + title + "</title>", false);
		
		//TODO change these to paths relative to a property, not the outputted HTML document
		addToOutput("<script src = 'Jscript/QTIV2JSCRIPT.js' type='text/javascript'></script>", false);
		addToOutput("<script src='Jscript/prototype.js' type='text/javascript'></script>", false);
		addToOutput("<script src='Jscript/scriptaculous.js' type='text/javascript'></script>", false);
		addToOutput("<link rel='stylesheet' href='style/r2q2.css' type='text/css' media='screen' />", false);
		
		//test for stylesheet
		Element ss = root.getChild("stylesheet", ns);
		if(ss != null){
			String style = "<link rel='stylesheet' ";
			style += attributesToString(ss);
			style += "/>";
			addToOutput(style, false);
		}
		
		//close the header
		decreaseDepth();
		addToOutput("</head>", false);
		
		//start the body
		Element itemBody = root.getChild("itemBody", ns);
		if(itemBody == null)
			throw new SpecificationViolationException("No itemBody element!");
		String bodyString = "<body ";
		bodyString += attributesToString(itemBody, attrBody);
		bodyString += " bgcolor='lightblue'>";
		addToOutput(bodyString);
		
		//output the title
		addToOutput("<h1>" + title + "</h1>", false);
		
		//wrap the page in a form
		addToOutput("<form  name = 'QuestionForm' method ='post' action='r2q2response.jsp' onsubmit='return doSubmit();' onreset='doReset();' > ");
		
		try {
			processElements(itemBody.getChildren());
		} catch (UnsupportedTagException e1) {
			throw new RenderingException(e1);
		}
		
		//add the submit/back/reset buttons
		Vector completionStatus = (Vector)((HashMap)vars[1].get("completionStatus")).get("completionStatus");
        addToOutput("<table>");
        addToOutput("<tr>");
        addToOutput("<td width = '100%' align = 'center'>");
        if(!completionStatus.elementAt(0).equals("completed") && (adaptive || (!adaptive && firstStage))){
        	addToOutput("<input name='submit' type='submit' value='SUBMIT ANSWER' />", false);
        	addToOutput("<input type='reset' value='RESET' />", false);
        }
        if(!firstStage)
        	addToOutput("<input name='submit' type='submit' value = 'GO BACK'/>", false);
        decreaseDepth();
        addToOutput("</td>", false);
        decreaseDepth();
        addToOutput("</tr>", false);
        decreaseDepth();
        addToOutput("</table>", false);
        
        //process any Modal feedback
        List<Content> modalFeedback = root.getChildren("modalFeedback", XHTMLRenderer.ns);
        if(modalFeedback != null && modalFeedback.size() > 0){
	        try {
	        	processElements(modalFeedback);
			} catch (UnsupportedTagException e1) {
				throw new RenderingException(e1);
			}
        }
        
		//close the form
		decreaseDepth();
		addToOutput("</form>", false);
		
		//close the body
		decreaseDepth();
		addToOutput("</body>", false);
		
		//close the html
		decreaseDepth();
		addToOutput("</html>", false);
		
		//return
		RendererResponse out_ = null;
		try {
			RenderingEngineResponse rer = new RenderingEngineResponse(variableNames, (HashMap)opts);
			out_ = new RendererResponse(rer, htmlString.getBytes("UTF-8"));
		} catch (UnsupportedEncodingException e) {
			throw new RenderingException(e);
		}
		return out_;
	}
	
	private void initialiseTagSets(){
		if(xhtmlTags == null){
			xhtmlTags = new HashSet<String>();
			xhtmlTags.add("a");
			xhtmlTags.add("abbr");
			xhtmlTags.add("acronym");
			xhtmlTags.add("address");
			xhtmlTags.add("b");
			xhtmlTags.add("big");
			xhtmlTags.add("blockquote");
			xhtmlTags.add("br");
			xhtmlTags.add("caption");
			xhtmlTags.add("cite");
			xhtmlTags.add("code");
			xhtmlTags.add("col");
			xhtmlTags.add("colgroup");
			xhtmlTags.add("dd");
			xhtmlTags.add("dfn");
			xhtmlTags.add("div");
			xhtmlTags.add("dl");
			xhtmlTags.add("dt");
			xhtmlTags.add("em");
			xhtmlTags.add("h1");
			xhtmlTags.add("h2");
			xhtmlTags.add("h3");
			xhtmlTags.add("h4");
			xhtmlTags.add("h5");
			xhtmlTags.add("h6");
			xhtmlTags.add("hr");
			xhtmlTags.add("i");
			xhtmlTags.add("img");
			xhtmlTags.add("kbd");
			xhtmlTags.add("li");
			xhtmlTags.add("object");
			xhtmlTags.add("ol");
			xhtmlTags.add("p");
			xhtmlTags.add("param");
			xhtmlTags.add("pre");
			xhtmlTags.add("q");
			xhtmlTags.add("samp");
			xhtmlTags.add("small");
			xhtmlTags.add("span");
			xhtmlTags.add("strong");
			xhtmlTags.add("sub");
			xhtmlTags.add("sup");
			xhtmlTags.add("table");
			xhtmlTags.add("tbody");
			xhtmlTags.add("td");
			xhtmlTags.add("tfoot");
			xhtmlTags.add("th");
			xhtmlTags.add("thead");
			xhtmlTags.add("tr");
			xhtmlTags.add("tt");
			xhtmlTags.add("ul");
			xhtmlTags.add("var");
		}
		if(attrBody == null){
			attrBody = new HashSet<String>();
			attrBody.add("id");
			attrBody.add("class");
		}
		if(interactionTags == null){
			interactionTags = new HashSet<String>();
			interactionTags.add("associableHotspot");
			interactionTags.add("associateInteraction");
			interactionTags.add("choiceInteraction");
			interactionTags.add("endAttemptInteraction");
			interactionTags.add("extendedTextInteraction");
			interactionTags.add("gap");
			interactionTags.add("gapImg");
			interactionTags.add("gapMatchInteraction");
			interactionTags.add("gapText");
			interactionTags.add("graphicAssociateInteraction");
			interactionTags.add("graphicGapMatchInteraction");
			interactionTags.add("graphicOrderInteraction");
			interactionTags.add("hotspotChoice");
			interactionTags.add("hotspotInteraction");
			interactionTags.add("hottext");
			interactionTags.add("hottextInteraction");
			interactionTags.add("inlineChoice");
			interactionTags.add("inlineChoiceInteraction");
			interactionTags.add("matchInteraction");
			interactionTags.add("orderInteraction");
			interactionTags.add("positionObjectInteraction");
			interactionTags.add("positionObjectStage");
			interactionTags.add("selectPointInteraction");
			interactionTags.add("simpleAssociableChoice");
			interactionTags.add("simpleChoice");
			interactionTags.add("simpleMatchSet");
			interactionTags.add("sliderInteraction");
			interactionTags.add("textEntryInteraction");
			interactionTags.add("uploadInteraction");
		}
		if(feedbackTags == null){
			feedbackTags = new HashSet<String>();
			feedbackTags.add("modalFeedback");
			feedbackTags.add("feedbackBlock");
			feedbackTags.add("feedbackInline");
		}
		if(templateTags == null){
			templateTags = new HashSet<String>();
			templateTags.add("templateBlock");
			templateTags.add("templateInline");
		}
	}
	
	public void processElement(Content _c) throws UnsupportedTagException, RenderingException, SpecificationViolationException{
		if(_c instanceof Text){
			Text t = (Text)_c;
			if(t.getTextTrim().length() > 0) addToOutput(t.getTextNormalize(), false);
		}
		else{
			Element e = (Element) _c;
			String elementName = e.getName();
			
			if(xhtmlTags.contains(elementName)){
				processHTML(e);
			}
			else if(interactionTags.contains(elementName)){
				processInteraction(e);
			}
			else if(feedbackTags.contains(elementName)
					|| templateTags.contains(elementName)){
				processHideable(e);
			}
			else if(elementName.equals("prompt")){
				addToOutput("<div class='prompt'>");
				processElements(e.getContent());
				decreaseDepth();
				addToOutput("</div>", false);
			}
			else if(elementName.equals("rubricBlock")){
				if(e.getAttribute("view").getValue().equals("candidate")){
					List<String> l = new ArrayList<String>();
					l.add("id");
					l.add("class");
					addToOutput("<div " + attributesToString(e, l) + ">");
					addToOutput("<h2>Instructions to candidate</h2>", false);
					processElements(e.getContent());
					decreaseDepth();
					addToOutput("</div>", false);
					addToOutput("<hr />", false);
				}
			}
			else if(elementName.equals("printedVariable")){//add the value of a variable to the output
				
				//try and find the variable
				String identifier = e.getAttribute("identifier").getValue();
				HashMap var = (HashMap)vars[1].get(identifier);
				if(var == null) var = (HashMap)vars[2].get(identifier);
				
				//if the variable exists
				if(var != null){
					//only output if it is single cardinality
					String cardinality = (String)var.get("cardinality");
					if(!cardinality.equals("single")) 
						throw new SpecificationViolationException(identifier + " is not single cardinality");
					
					//only output if it has a value
					String value = (String)((Vector)var.get(identifier)).get(0);
					if(value != null) addToOutput(value, false);
				}
				else 
					throw new SpecificationViolationException("Cannot print non-existent variable " + identifier);
			}
			else{
				throw new UnsupportedTagException(elementName);
			}
		}
	}
	
	public void processElements(List<Content> _e) throws UnsupportedTagException, RenderingException, SpecificationViolationException{
		Iterator<Content> iter = _e.iterator();
		processElements(iter);
	}
	
	public void processElements(Iterator<Content> _iter) throws UnsupportedTagException, RenderingException, SpecificationViolationException{
		while(_iter.hasNext()) processElement(_iter.next());
	}
	
	void processHTML(Element _e) throws UnsupportedTagException, RenderingException, SpecificationViolationException{
		String temp = "<" + _e.getName() + " ";
		temp += attributesToString(_e);
		
		//process any sub-nodes
		List<Content> children = _e.getContent();
		if(children.size() > 0){
			temp += ">";
			addToOutput(temp);
			processElements(children);
			decreaseDepth();
			addToOutput("</" + _e.getName() + ">", false);
		}
		else{
			temp += "/>";
			addToOutput(temp, false);
		}
	}
	
	void processInteraction(Element _e) throws UnsupportedTagException, RenderingException, SpecificationViolationException{
		String name = _e.getName();
		boolean newIntProc = false;
		boolean oldIntProc = currentIntProc != null;
		
		if(!interactionTags.contains(name)) throw new UnsupportedTagException(name + " in interaction processing");
		
		//interactions not allowed within feedback on non-adaptive questions
		if(!adaptive && (inTag.contains("feedbackBlock") || inTag.contains("feedbackInline")))
			throw new SpecificationViolationException("Cannot have interactions below feedback in a non-adaptive item");
		//interactions not allowed within modal feedback on any question
		if(adaptive && inTag.contains("modalFeedback"))
			throw new SpecificationViolationException("Cannot have interactions below modal feedback");
		//interactions not allowed within template blocks
		if(inTag.contains("templateBlock") || inTag.contains("templateInline"))
			throw new SpecificationViolationException("Cannot have interactions below templates");
		
		String elementName = _e.getName();
		
		//TODO use reflection
		if(elementName.equals("choiceInteraction")){
			currentIntProc = new ChoiceInteractionProcessor(this);
			newIntProc = true;
		}
		else if(elementName.equals("textEntryInteraction")){
			currentIntProc = new TextEntryInteractionProcessor(this);
			newIntProc = true;
		}
		else if(elementName.equals("associateInteraction")){
			currentIntProc = new AssociateInteractionProcessor(this);
			newIntProc = true;
		}
		else if(elementName.equals("sliderInteraction")){
			currentIntProc = new SliderInteractionProcessor(this);
			newIntProc = true;
		}
		else if(elementName.equals("inlineChoiceInteraction")){
			currentIntProc = new InlineChoiceInteractionProcessor(this);
			newIntProc = true;
		}
		else if(elementName.equals("selectPointInteraction")){
			currentIntProc = new SelectPointInteractionProcessor(this);
			newIntProc = true;
		}
		else if(elementName.equals("hotspotInteraction")){
			currentIntProc = new HotspotInteractionProcessor(this);
			newIntProc = true;
		}
		else if(elementName.equals("matchInteraction")){
			currentIntProc = new MatchInteractionProcessor(this);
			newIntProc = true;
		}
		else if(elementName.equals("graphicOrderInteraction")){
			currentIntProc = new GraphicOrderInteractionProcessor(this);
			newIntProc = true;
		}
		else if(elementName.equals("graphicAssociateInteraction")){
			currentIntProc = new GraphicAssociateInteractionProcessor(this);
			newIntProc = true;
		}
		else if(elementName.equals("graphicGapMatchInteraction")){
			currentIntProc = new GraphicGapMatchInteractionProcessor(this);
			newIntProc = true;
		}
		else if(elementName.equals("orderInteraction")){
			currentIntProc = new OrderInteractionProcessor(this);
			newIntProc = true;
		}
		else if(elementName.equals("hottextInteraction")){
			currentIntProc = new HottextInteractionProcessor(this);
			newIntProc = true;
		}
		else if(elementName.equals("positionObjectStage")){
			currentIntProc = new PositionObjectInteractionProcessor(this);
			newIntProc = true;
		}
		else if(elementName.equals("gapMatchInteraction")){
			currentIntProc = new GapMatchInteractionProcessor(this);
			newIntProc = true;
		}
		
		//TODO wasteful creating a new object
		if(oldIntProc && newIntProc)
			throw new SpecificationViolationException("Can't have nested interactions");
		
		if(currentIntProc != null){
			currentIntProc.process(_e);
			if(newIntProc) currentIntProc = null;
		}
	}
	
	void processHideable(Element _e) throws UnsupportedTagException, RenderingException, SpecificationViolationException{
		//TODO see bug 19
		//if(!firstStage){
			String name = _e.getName();
			inTag.add(name);
			
			boolean isFeedback = feedbackTags.contains(name);
			boolean isTemplate = templateTags.contains(name);
			
			if(!isFeedback && !isTemplate) 
				throw new UnsupportedTagException(name + " not a feedback tag");
			
			HashMap hideVars = null;
			String varIdentifier = null;
			
			//extract the common attributes
			if(isFeedback){
				varIdentifier = _e.getAttributeValue("outcomeIdentifier");
				hideVars = (HashMap)vars[1].get(varIdentifier);
			}
			else{
				varIdentifier = _e.getAttributeValue("templateIdentifier");
				hideVars = (HashMap)vars[2].get(varIdentifier);
			}
			String identifier = _e.getAttributeValue("identifier");
			String showHide = _e.getAttributeValue("showHide");
			
			//determine whether to show or hide
			Vector<String> values = (Vector)hideVars.get(varIdentifier);
			if(values == null) System.out.println("Values for identifier " + varIdentifier + " are null");
			else {
				System.out.print("Values for identifier " + varIdentifier + ": ");
				for(String value: values) System.out.print(value);
				System.out.println();
			}
			
			boolean match = false;
			if(values != null) match = values.contains(identifier);
			if((match && showHide.equals("show"))
					|| (!match && showHide.equals("hide"))){
				//display the feedback/template
				if(name.equals("modalFeedback") && !firstStage){
					addToOutput("<hr />", false);
					addToOutput("<div class='r2q2ModalFeedback'>");
					addToOutput("<h2>Feedback</h2>", false);
					Attribute attr = _e.getAttribute("title");
					if(attr != null){
						addToOutput("<h2>" + attr.getValue() + "</h2>", false);
					}
					processElements(_e.getContent());
					decreaseDepth();
					addToOutput("</div>", false);
				}
				else if(name.equals("feedbackBlock")){
					List<String> l = new ArrayList<String>();
					l.add("id");
					l.add("class");
					addToOutput("<div " + attributesToString(_e, l) + ">");
					processElements(_e.getContent());
					decreaseDepth();
					addToOutput("</div>", false);
				}
				else if(name.equals("feedbackInline")){
					List<String> l = new ArrayList<String>();
					l.add("id");
					l.add("class");
					addToOutput("<b><span " + attributesToString(_e, l) + ">");
					processElements(_e.getContent());
					decreaseDepth();
					addToOutput("</span></b>", false);
				}
			}
			
			inTag.remove(name);
		//}
	}
	
	String attributesToString(Element _e, Collection<String> _c){
		List<Attribute> attrList = _e.getAttributes();
		String out_ = "";
		Iterator<Attribute> iter = attrList.iterator();
		while(iter.hasNext()){
			Attribute attr = iter.next();
			if(_c == null || _c.contains(attr.getName())){
				out_ += attr.getName() + "=\"" + attr.getValue() + "\" ";
			}
		}
		return out_;
	}
	
	String attributesToString(Element _e){
		return attributesToString(_e, null);
	}
	
	public void addToOutput(String _in, boolean _increase){
		htmlString += depth + _in + "\n";
		if(_increase) increaseDepth();
	}
	
	public void addToOutput(String _in){
		addToOutput(_in, true);
	}
	
	public void increaseDepth(){
		depth += "\t";
	}
	
	public void decreaseDepth(){
		depth = depth.substring(0, depth.length() - 1);
	}
	

	//TODO DELETEME
	public static void main(String args[]) throws Throwable{
		try{
			XHTMLRenderer x = new XHTMLRenderer();
			byte[] a = StreamCopier.copyToByteArray(new FileInputStream( new File ("qti/gap_match.xml")));
			
			InitialisationEngine ie = new InitialisationEngine();
			HashMap[] hm = ie.intialiseQuestion(new String(a));
			RendererResponse r = x.render(a, hm, null, null, false);
			System.out.println(new String(r.getContent()));
		}
		catch (Throwable t){
			System.err.println(t.getMessage());
			t.printStackTrace();
		}
	}
}
