/*
<LICENCE>

Copyright (c) 2007, University of Southampton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this
	list of conditions and the following disclaimer.

  *	Redistributions in binary form must reproduce the above copyright notice,
	this list of conditions and the following disclaimer in the documentation
	and/or other materials provided with the distribution.

  *	Neither the name of the University of Southampton nor the names of its
	contributors may be used to endorse or promote products derived from this
	software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

</LICENCE>
*/

package r2q2.rendering.xhtml.processors;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import org.jdom.Attribute;
import org.jdom.Content;
import org.jdom.Element;

import com.gavinwillingham.util.properties.PropertiesLoader;

import r2q2.SpecificationViolationException;
import r2q2.rendering.RenderingException;
import r2q2.rendering.xhtml.UnsupportedTagException;
import r2q2.rendering.xhtml.XHTMLRenderer;
import r2q2.rendering.xhtml.processors.util.ListElements;
import r2q2.rendering.xhtml.structure.xhtml.XHTMLObject;

public class AssociateInteractionProcessor extends InteractionProcessor {

	private boolean shuffle = false;
	private int maxAssociations = 1;
	private String basePath = null;
	private int choiceSize = 0;
	private int choiceNum = 0;
	
	public AssociateInteractionProcessor(XHTMLRenderer re) throws RenderingException {
		super(re);
		
		try {
			basePath = PropertiesLoader.getProperty("r2q2.router.content.httpBasePath");
		} catch (Exception e) {
			throw new RenderingException("Error loading htmlBasePath property", e);
		}
		
		supportedTags.add("associateInteraction");
		supportedTags.add("prompt");
		supportedTags.add("simpleAssociableChoice");
	}

	@Override
	public void process(Element _e) throws UnsupportedTagException,
			RenderingException, SpecificationViolationException {
		
		String elementName = _e.getName();
		
		//test valid element
		if(!supportedTags.contains(elementName))
			throw new UnsupportedTagException(elementName + "not supported in AssociateInteractionProcessor");
		
		if(elementName.equals("associateInteraction")){
			processAssociateInteraction(_e);
		}
		else if(elementName.equals("simpleAssociableChoice")){
			processSimpleAssociableChoice(_e);
		}
		else if(elementName.equals("prompt")){
			engine.processElement(_e);
		}
	}
	
	private void processAssociateInteraction(Element _e) throws UnsupportedTagException, RenderingException, SpecificationViolationException {
		
    	String attname = "";
		try{
			//Now process all of the attributes in the array 
	        Iterator<Attribute> attrs = _e.getAttributes().iterator();
	        while (attrs.hasNext()){          
	            Attribute attr = attrs.next();                  
	            attname = attr.getName();                  
	            String value = attr.getValue();  
	            if(attname.equals("responseIdentifier"))
	                responseIdentifier = value;
	            else if(attname.equals("maxAssociations"))
	                maxAssociations = Integer.parseInt(value);
	            else if(attname.equals("shuffle"))
	                shuffle = Boolean.parseBoolean(value);
	        }
		}
	    catch(NumberFormatException e){
	    	throw new SpecificationViolationException(attname + " could not be converted to the appropriate type", e);
	    }
		
        //test that response variable is identifier
        HashMap<String, HashMap> responseVars = engine.vars[0];
    	HashMap interactionResponse = responseVars.get(responseIdentifier);
    	if(interactionResponse == null)
    		throw new SpecificationViolationException("ResponseVar " + responseIdentifier + " not present");
    	String baseType = (String) interactionResponse.get("baseType");
    	if(!baseType.equals("pair"))
    		throw new SpecificationViolationException("Cannot assign associations to a variable (" + responseIdentifier + ") that is not of base type pair in associateInteraction");
    	String cardinality = (String) interactionResponse.get("cardinality");
    	if(!cardinality.equals("single") && !cardinality.equals("multiple"))
    		throw new SpecificationViolationException("Cannot assign associations to a variable (" + responseIdentifier + ") that is not single or multiple cardinality in associateInteraction");
    	if(cardinality.equals("single") && maxAssociations != 1)
    		throw new SpecificationViolationException("Cannot assign multiple associations to a variable (" + responseIdentifier + ") that is not multiple cardinality in associateInteraction");
    	
    	//get any previous values, i.e. user response
    	response = (Vector) interactionResponse.get("candidateValues");
		
		//output the prompt
    	Element prompt = _e.getChild("prompt", XHTMLRenderer.ns);
    	if(prompt != null) engine.processElement(prompt);
    	
    	//get the choices
    	List<Element> choices = _e.getChildren("simpleAssociableChoice", engine.ns);
    	if(choices == null || choices.size() < 1)
    		throw new SpecificationViolationException("No choices beneath associationInteraction");
    	choiceSize = choices.size();
	    
    	if(shuffle){
        	List orderList = engine.opts.get(responseIdentifier);
        	if(engine.opts.get(responseIdentifier) == null)
        	{
        		choices = ListElements.randomiseList(choices);
        		//set the hashtable
        		List<String> map = ListElements.getOrder(choices);
        		List<List<String>> mapList= new Vector<List<String>>();
        		mapList.add(map);
        		engine.opts.put(responseIdentifier, mapList);
        	}else
        	{
        		//need to fetch hashtable
        		choices = ListElements.orderList(choices, (List<String>)orderList.get(0));
        	}
        }
    	
	    //start the applet
	    engine.addToOutput("<applet"); 
        engine.addToOutput("codebase='" + basePath + "'", false);
    	engine.addToOutput("code = 'BoundedGraphicalApplet.class'", false);
    	engine.addToOutput("width = '" + (choiceSize * 60) + "'", false);
    	engine.addToOutput("height = '" + (choiceSize * 60) + "'", false);
    	engine.addToOutput("hspace = '0'", false);
    	engine.addToOutput("vspace = '0'", false);
    	engine.addToOutput("align = 'middle'", false);
    	engine.addToOutput(">", false);
    	engine.addToOutput("<param name ='identifier' value = '" + responseIdentifier + "'>", false);
    	engine.addToOutput("<param name ='baseType' value = 'pair'>", false);
    	engine.addToOutput("<param name ='operation_mode' value = 'graphic_associate_interaction'>", false);
    	engine.addToOutput("<param name ='number_of_responses' value ='" + maxAssociations + "'>", false);
    	
    	//process the choices
    	engine.addToOutput("<param name ='hotspot_count' value ='" + choiceSize + "'>", false);
    	process(choices);
    	
    	//add any previous selections
    	if(response != null && response.size() > 0){
    		String feedback = "";
    		int i=0;
    		for(; i<response.size() - 1; i++){
    			feedback += response.get(i) + ",";
    		}
    		feedback += response.get(i);
    		engine.addToOutput("<param name='feedback' value='" + feedback + "' />", false);
    	}
    	
    	//close the applet
    	engine.decreaseDepth();
    	engine.addToOutput("</applet>", false);
	}

	private void processSimpleAssociableChoice(Element _e) throws SpecificationViolationException {
		//extract the attributes
		String id = _e.getAttributeValue("identifier");
		if(id == null)
			throw new SpecificationViolationException("No identifier attribute for simpleAssociableChoice");
		String matchGroup = _e.getAttributeValue("matchGroup");
		
		//build the output
		String temp = "<param name='hotspot" +choiceNum++ + "' " +
				"value='" + id + "" +
				"::" + _e.getTextNormalize(); 
		
		if(matchGroup != null){
			String[] matches = matchGroup.split(",");
			for(int i=0; i<matches.length; i++)
				temp += "::" + matches[i];
		}
		
		temp += "' />";
				
		engine.addToOutput(temp, false);
	}

}

