/*
<LICENCE>

Copyright (c) 2007, University of Southampton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this
	list of conditions and the following disclaimer.

  *	Redistributions in binary form must reproduce the above copyright notice,
	this list of conditions and the following disclaimer in the documentation
	and/or other materials provided with the distribution.

  *	Neither the name of the University of Southampton nor the names of its
	contributors may be used to endorse or promote products derived from this
	software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

</LICENCE>
*/

package r2q2.rendering.xhtml.processors;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import org.jdom.*;

import r2q2.SpecificationViolationException;
import r2q2.rendering.Renderer;
import r2q2.rendering.RenderingException;
import r2q2.rendering.xhtml.UnsupportedTagException;
import r2q2.rendering.xhtml.XHTMLRenderer;
import r2q2.rendering.xhtml.processors.util.ListElements;

public class ChoiceInteractionProcessor extends InteractionProcessor {
	
	//interaction settings
	private int maxChoices = 0;
	private boolean shuffle = false;
	private Vector response = null;
	
	public ChoiceInteractionProcessor(XHTMLRenderer re) throws RenderingException {
		super(re);
		supportedTags.add("choiceInteraction");
		supportedTags.add("simpleChoice");
		supportedTags.add("prompt");
	}
	
	public void process(Element _e) throws UnsupportedTagException, RenderingException, SpecificationViolationException {
		
		String elementName = _e.getName();
		
		if(!supportedTags.contains(elementName)){
			throw new UnsupportedTagException(elementName + " in choiceInteraction");
		}
		else if(_e.getName().equals("choiceInteraction")){
			try {
				processChoiceInteraction(_e);
			} catch (DataConversionException e1) {
				throw new RenderingException(e1);
			}
		}
		else if(_e.getName().equals("simpleChoice")){
			processSimpleChoice(_e);
		}
		else if(_e.getName().equals("prompt")){
			processPrompt(_e);
		}
	}
	
	private void processChoiceInteraction(Element _e) throws UnsupportedTagException, DataConversionException, RenderingException, SpecificationViolationException{
		
		String attname = ""; 
		try{
			//Now process all of the attributes in the array 
	        Iterator<Attribute> attrs = _e.getAttributes().iterator();
	        while (attrs.hasNext()){          
	            Attribute attr = attrs.next();                  
	            attname = attr.getName();                  
	            String value = attr.getValue();  
	            if(attname.equals("responseIdentifier"))
	                responseIdentifier = value;
	            else if(attname.equals("maxChoices"))
	                maxChoices = Integer.parseInt(value);
	            else if(attname.equals("shuffle"))
	                shuffle = Boolean.parseBoolean(value);
	        }
	    }
	    catch(NumberFormatException e){
	    	throw new SpecificationViolationException(attname + " could not be converted to the appropriate type", e);
	    }
        
        //test that response variable is identifier
	    if(engine.vars == null) System.out.println("engine.vars is null");//TODO delete me
        HashMap<String, HashMap> responseVars = engine.vars[0];
    	HashMap interactionResponse = responseVars.get(responseIdentifier);
    	if(interactionResponse == null)
    		throw new SpecificationViolationException("ResponseVar " + responseIdentifier + " not present");
    	String baseType = (String) interactionResponse.get("baseType");
    	if(!baseType.equals("identifier"))
    	{
    		//TODO delete
    		System.out.println("response type is : "+baseType);
    		throw new SpecificationViolationException("Cannot assign choices to a variable (" + responseIdentifier + ") that is not of base type identifier");
    	}
    	//TODO delete
    	System.out.println("values type is " + interactionResponse.get("candidateValues"));
    	response = (Vector) interactionResponse.get("candidateValues");
    	
        //test that response variable is multiple cardinality if maxChoices !=1
        if(maxChoices != 1){
	        String cardinality = (String) interactionResponse.get("cardinality");
	    	if(cardinality.equals("single"))
	    		throw new SpecificationViolationException("Cannot assign multiple choices to a single cardinality variable (" + responseIdentifier + ")");
	    }
        
        //order the choices
        List<Element> children = _e.getChildren();
        Element prompt = _e.getChild("prompt", engine.ns);
        if(prompt != null){
	        children.remove(prompt);
	        process(prompt);
        }
        
        if(children == null || children.size() == 0)
        	throw new SpecificationViolationException("No children of choiceInteraction element");
        
        //if necessary, shuffle the order of the choices
        if(shuffle){
        	List orderList = engine.opts.get(responseIdentifier);
        	if(engine.opts.get(responseIdentifier) == null)
        	{
        		children = ListElements.randomiseList(children);
        		//set the hashtable
        		List<String> map = ListElements.getOrder(children);
        		List<List<String>> mapList= new Vector<List<String>>();
        		mapList.add(map);
        		engine.opts.put(responseIdentifier, mapList);
        	}else
        	{
        		//need to fetch hashtable
        		children = ListElements.orderList(children, (List<String>)orderList.get(0));
        	}
        }
        
        //start the table
        engine.addToOutput("<table border='border'>");
        
        //process the children
        engine.processElements((List<Content>)(List)children);
        
        //close the table
        engine.decreaseDepth();
        engine.addToOutput("</table>", false);
	}
	
	private void processSimpleChoice(Element _e) throws UnsupportedTagException, RenderingException, SpecificationViolationException{
		//Start the table row
        engine.addToOutput("<tr align = 'center'>");
        
        //output the answer text
        engine.addToOutput("<td align = 'center'>");
        engine.processElements(_e.getContent());
        engine.decreaseDepth();
        engine.addToOutput("</td>", false);
                
        //add the option input
        String identifier = _e.getAttribute("identifier").getValue();
    	String temp = "";
        
    	temp = "<input type='";
        if(maxChoices == 1) temp += "radio";
        else{
        	temp += "checkbox";	
        	if(maxChoices > 1)
        		temp += "' onclick='limitChecks(" + maxChoices + ", \"" + responseIdentifier + "\", \"" + identifier + "\")";
        }
        temp += "' name='" + responseIdentifier + "' value='" + identifier + "'";
        if(response != null && response.contains(identifier)) temp += " checked='checked'";
        temp += "/>";
        
        engine.addToOutput("<td>");
        engine.addToOutput(temp, false);
        engine.decreaseDepth();
        engine.addToOutput("</td>", false);
        
        //close the row
        engine.decreaseDepth();
        engine.addToOutput("</tr>", false);
	}
	
	private void processPrompt(Element _e) throws UnsupportedTagException, RenderingException, SpecificationViolationException{
		engine.addToOutput("<tr align='center'>");
		engine.addToOutput("<td colspan='2'>");
		engine.processElements(_e.getContent());
		engine.decreaseDepth();
		engine.addToOutput("</td>", false);
		engine.decreaseDepth();
		engine.addToOutput("</tr>", false);
	}
}