/*
<LICENCE>

Copyright (c) 2007, University of Southampton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this
	list of conditions and the following disclaimer.

  *	Redistributions in binary form must reproduce the above copyright notice,
	this list of conditions and the following disclaimer in the documentation
	and/or other materials provided with the distribution.

  *	Neither the name of the University of Southampton nor the names of its
	contributors may be used to endorse or promote products derived from this
	software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

</LICENCE>
*/

package r2q2.rendering.xhtml.processors;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import org.jdom.Attribute;
import org.jdom.Content;
import org.jdom.Element;

import r2q2.SpecificationViolationException;
import r2q2.rendering.RenderingException;
import r2q2.rendering.xhtml.UnsupportedTagException;
import r2q2.rendering.xhtml.XHTMLRenderer;
import r2q2.rendering.xhtml.processors.util.ListElements;
import r2q2.util.lists.ListUtils;

public class GapMatchInteractionProcessor extends InteractionProcessor {
	
	private int gapCount = 1;
	private boolean shuffle = false;
	private HashMap<String, Vector<String>> matchGroups = new HashMap<String, Vector<String>>();
	private List<String> gaps = new Vector<String>();
	private List<String> gapChoices = new Vector<String>();
	private HashMap<String, Integer> maxChoices = new HashMap<String, Integer>();

	public GapMatchInteractionProcessor(XHTMLRenderer re) throws RenderingException {
		super(re);
		
		supportedTags.add("gapMatchInteraction");
		supportedTags.add("gap");
		supportedTags.add("gapText");
		supportedTags.add("gapImg");
	}

	@Override
	public void process(Element _e) throws UnsupportedTagException,
			RenderingException, SpecificationViolationException {

		String elementName = _e.getName();
		
		//test valid element
		if(!supportedTags.contains(elementName))
			throw new UnsupportedTagException(elementName + "not supported in MatchInteractionProcessor");
		
		if(elementName.equals("gapMatchInteraction")){
			processGapMatchInteraction(_e);
		}
		else if(elementName.equals("gap")){
			processGap(_e);
		}
		else if(elementName.equals("gapText") || elementName.equals("gapImg")){
			processGapChoice(_e);
		}

	}

	private void processGapMatchInteraction(Element _e) throws SpecificationViolationException, UnsupportedTagException, RenderingException {

		List<Attribute> attrs = _e.getAttributes();
        //Now process all of the attributes in the array. 
        String attname = null;
        try{
	        for (Attribute attr: attrs){                   
	            attname = attr.getName();                  
	            String value = attr.getValue();  
	            if(attname.equals("responseIdentifier"))
	                responseIdentifier = value;
	            if(attname.equals("shuffle"))
	                shuffle = Boolean.parseBoolean(value);
	        }  
        }catch(NumberFormatException nfe)
        {
        	throw new SpecificationViolationException(attname + " could not be converted to the appropriate type");
        }
        
        HashMap<String, HashMap> responseVars = engine.vars[0];
    	HashMap interactionResponse = responseVars.get(responseIdentifier);
        response = (Vector) interactionResponse.get("candidateValues");
        
        //grab the child nodes
        List<Content> content = _e.getContent();
        List<Element> gapList = _e.getChildren("gapText", XHTMLRenderer.ns);
        List<Element> gapImg = _e.getChildren("gapImg", XHTMLRenderer.ns);
        gapList.addAll(gapImg);
       
        //remove the gapImg/Text nodes
        List tempGaps = ListUtils.cloneList(gapList);
        content.removeAll(gapList);
        gapList = tempGaps;
		
        //process the body/prompt
        engine.processElements(content);
		
        //shuffle the gapChoices
        if(shuffle){
        	List orderList = engine.opts.get(responseIdentifier);
        	if(engine.opts.get(responseIdentifier) == null)
        	{
        		gapList = ListElements.randomiseList(gapList);
        		//set the hashtable
        		List<String> map = ListElements.getOrder(gapList);
        		List<List<String>> mapList= new Vector<List<String>>();
        		mapList.add(map);
        		engine.opts.put(responseIdentifier, mapList);
        	}else
        	{
        		//need to fetch hashtable
        		gapList = ListElements.orderList(gapList, (List<String>)orderList.get(0));
        	}
        }
        
        //generate the table
        engine.addToOutput("<table>");
        
        //build the top row (choices)
        engine.addToOutput("<tr>");
        engine.addToOutput("<td></td>", false);
        process(gapList);
        engine.decreaseDepth();
        engine.addToOutput("</tr>", false);
        
        if(gapChoices.size() < 1)
        	throw new SpecificationViolationException("No gapchoices below gapmatchinteraction");
        if(gaps.size() < 1)
        	throw new SpecificationViolationException("No gaps below gapmatchinteraction");
        
        //add the gaps
        int i=1;
        for(String nextGap: gaps){
        	engine.addToOutput("<tr>");
        	engine.addToOutput("<td>");
        	engine.addToOutput("GAP " + i++, false);
        	engine.decreaseDepth();
        	engine.addToOutput("</td>", false);
        	
        	//add the checkboxes
        	for(String nextChoice: gapChoices){
	        	engine.addToOutput("<td>");
	        	
	        	//determine whether this choice is allowed or not
	        	String disabled = "";
	        	Vector<String> nextGroup1 = matchGroups.get(nextChoice);
	        	Vector<String> nextGroup2 = matchGroups.get(nextGap);
	        	if((nextGroup1 != null && !nextGroup1.contains(nextGap))
	        			&& (nextGroup2 != null && !nextGroup2.contains(nextChoice)))
	        		disabled = " disabled='disabled'";
	        	
	        	//set the boxes that should be check from the previous stage
	        	String checked = "";
	        	if(response != null){
	        		for(String next: response){
	        			String[] splitPair = next.split(" ");
	        			if(splitPair.length == 2){
	        				if(splitPair[0].equals(nextChoice) 
	        					&& splitPair[1].equals(nextGap))
	        						checked = "checked='checked'";
	        			}
	        		}
	        	}
	        	
	        	engine.addToOutput("<input name='" + responseIdentifier + "[]' type='checkbox' value='" + nextChoice + " " + nextGap +"' onclick='checkGapMatch(\"" + responseIdentifier + "\", this)'" + disabled + checked + " />", false);
	        	engine.decreaseDepth();
	        	engine.addToOutput("</td>", false);
        	}
        	
        	engine.decreaseDepth();
        	engine.addToOutput("</tr>", false);
        }
        
        //close the table
        engine.decreaseDepth();
        engine.addToOutput("</table>", false);
        
        //output the script-call code to set up match limiting
        engine.addToOutput("<script type='text/javascript'>");
        engine.addToOutput("//<![CDATA[");
        
        String temp = "setupGapMatch('" + responseIdentifier + "', {";
        
        Iterator<String> iter = gapChoices.iterator();
        while(true){
        	String nextKey = iter.next();
        	Integer nextValue = maxChoices.get(nextKey);
        	temp += nextKey + ":" + nextValue.toString();
        	if(!iter.hasNext()) break;
        	temp += ",";
        }
        
        temp += "}, {";
        
        iter = gaps.iterator();
        while(true){
        	String nextKey = iter.next();
        	temp += nextKey + ":1";
        	if(!iter.hasNext()) break;
        	temp += ",";
        }
        
        temp += "});";
        engine.addToOutput(temp, false);
        
        engine.decreaseDepth();
        engine.addToOutput("//]]>", false);
        engine.decreaseDepth();
        engine.addToOutput("</script>", false);
	}

	private void processGap(Element _e) throws SpecificationViolationException {
		
		//grab the attributes
		String identifier = _e.getAttributeValue("identifier");
		if(identifier == null)
			throw new SpecificationViolationException("No identifier attribute for gap");
		gaps.add(identifier);
		
		String matchGroup = _e.getAttributeValue("matchGroup");
		if(matchGroup != null){
			//store the match groups
			String[] groups= matchGroup.split(" ");
			Vector<String> tmp = new Vector<String>();
			for(int i=0; i<groups.length; i++)
				tmp.add(groups[i]);
			matchGroups.put(identifier, tmp);
		}
		
		//generate the output
		engine.addToOutput("<span class='gap' id='" + responseIdentifier + ":" + identifier + "'>GAP " + gapCount++ + "</span>", false);
	}
	
	private void processGapChoice(Element _e) throws SpecificationViolationException {
		
		//do processing common to getText and gapImg
		String identifier = _e.getAttributeValue("identifier");
		if(identifier == null)
			throw new SpecificationViolationException("No identifier specified in gapChoice");
		gapChoices.add(identifier);
		
		String matchMax = _e.getAttributeValue("matchMax");
		if(matchMax == null)
			throw new SpecificationViolationException("No matchMax specified");
		Integer matchMaxInt;
		try{
			matchMaxInt = Integer.parseInt(matchMax);
		}
		catch (NumberFormatException e){
			throw new SpecificationViolationException("matchMax specified but not integer");
		}
		
		maxChoices.put(identifier, new Integer(matchMaxInt));
		
		String matchGroup = _e.getAttributeValue("matchGroup");
		if(matchGroup != null){
			//store the match groups
			String[] groups= matchGroup.split(" ");
			Vector<String> tmp = new Vector<String>();
			for(int i=0; i<groups.length; i++)
				tmp.add(groups[i]);
			matchGroups.put(identifier, tmp);
		}
		
		engine.addToOutput("<td id='" + responseIdentifier + ":" + identifier + "'>");
		engine.addToOutput(_e.getTextNormalize(), false);
		engine.decreaseDepth();
		engine.addToOutput("</td>", false);
	}
}
