/*
<LICENCE>

Copyright (c) 2007, University of Southampton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this
	list of conditions and the following disclaimer.

  *	Redistributions in binary form must reproduce the above copyright notice,
	this list of conditions and the following disclaimer in the documentation
	and/or other materials provided with the distribution.

  *	Neither the name of the University of Southampton nor the names of its
	contributors may be used to endorse or promote products derived from this
	software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

</LICENCE>
*/

package r2q2.rendering.xhtml.processors;

import java.awt.Toolkit;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Vector;
import java.util.List;

import javax.imageio.ImageIO;

import org.jdom.Attribute;
import org.jdom.Element;

import com.gavinwillingham.util.properties.NoSuchPropertyException;
import com.gavinwillingham.util.properties.PropertiesLoader;

import r2q2.SpecificationViolationException;
import r2q2.rendering.RenderingException;
import r2q2.rendering.xhtml.UnsupportedTagException;
import r2q2.rendering.xhtml.XHTMLRenderer;
import r2q2.rendering.xhtml.processors.util.ObjectProcessor;
import r2q2.rendering.xhtml.structure.xhtml.XHTMLObject;
import r2q2.util.strings.TokenUtils;

/**
 * 
 * TODO javadoc
 *
 * @author Alex Frazer
 *
 */
public class PositionObjectInteractionProcessor extends GraphicInteractionProcessor {
	
HashSet<String> supportedObjectTypes = null;
	
	private String responseIdentifier = "";
	private int maxChoices = 1;
	private Vector<String> response = null;
	private Vector candVals;

	public PositionObjectInteractionProcessor(XHTMLRenderer re) throws RenderingException {
		super(re);
		
		candVals = new Vector();
		
		supportedTags.add("positionObjectInteraction");
		supportedTags.add("positionObjectStage");
		supportedTags.add("object");
		
		loadSupportedObjectTypes();
	}
	
	private void loadSupportedObjectTypes() throws RenderingException {
		if(supportedObjectTypes == null){
			//load the types from properties
			try {
				String props = PropertiesLoader.getProperty("r2q2.rendering.xhtml.graphicInteraction.supportedObjectTypes");
				supportedObjectTypes = TokenUtils.tokensToHashSet(props, " ");
			} catch (IOException e) {
				throw new RenderingException("Couldn't load properties file in positionObjectInteraction!");
			} catch (NoSuchPropertyException e) {
				throw new RenderingException("Couldn't find property in positionObjectInteraction!");
			}
		}
	}
	
	public void process(Element child) throws UnsupportedTagException, RenderingException, SpecificationViolationException {
		
		String elementName = child.getName();
		
		//test valid tag
		if(!supportedTags.contains(elementName))
			throw new UnsupportedTagException(elementName + " in positionObjectInteraction");
		
		if(elementName.equals("positionObjectStage"))
			processPositionObjectStage(child);
	}
	
	private void processPositionObjectStage(Element _e) throws SpecificationViolationException, RenderingException, UnsupportedTagException{
		
		// output the prompt
    	Element prompt = _e.getChild("prompt", XHTMLRenderer.ns);
    	if(prompt != null) processPrompt(prompt);
    	
    	//find the object node for the main image
    	Element object = _e.getChild("object", XHTMLRenderer.ns);
    	
    	if(object == null)
    		throw new SpecificationViolationException("No object found under SelectPointInteraction");
    	
    	XHTMLObject obj = processObject(object);
    	
    	// there will now be one or more positionObjectInteractions - process them now!
    	List interactions = _e.getChildren("positionObjectInteraction", XHTMLRenderer.ns);
    	
    	// set up a Vector to contain the different interactions
    	int len = interactions.size();          
        Vector ints = new Vector();
    	
    	//Iterate on NodeList of child nodes.          
    	for (int i = 0; i < len; i++) {               
    		Element interaction = (Element)interactions.get(i);
    		
    		String attname = "";
    		
    		try{
    			//Now process all of the attributes in the array 
    	        Iterator<Attribute> attrs = interaction.getAttributes().iterator();
    	        while (attrs.hasNext()){          
    	            Attribute attr = attrs.next();                  
    	            attname = attr.getName();                  
    	            String value = attr.getValue();  
    	            if(attname.equals("responseIdentifier"))
    	                responseIdentifier = value;
    	            else if(attname.equals("maxChoices"))
    	                maxChoices = Integer.parseInt(value);
    	        }
    		}
    	    catch(NumberFormatException e){
    	    	throw new SpecificationViolationException(attname + " could not be converted to the appropriate type", e);
    	    }
    	    
    	    // test that response variable is identifier
            HashMap<String, HashMap> responseVars = engine.vars[0];
        	HashMap interactionResponse = responseVars.get(responseIdentifier);
        	if(interactionResponse == null)
        		throw new SpecificationViolationException("ResponseVar " + responseIdentifier + " not present");
        	String baseType = (String) interactionResponse.get("baseType");
        	if(!baseType.equals("point"))
        		throw new SpecificationViolationException("Cannot assign points to a variable (" + responseIdentifier + ") that is not of base type point");
        	String cardinality = (String) interactionResponse.get("cardinality");
        	if(!cardinality.equals("single") && !cardinality.equals("multiple"))
        		throw new SpecificationViolationException("Cannot assign match interaction to a variable (" + responseIdentifier + ") that is not single or multiple cardinality");
        	
            //test that response variable is multiple cardinality if maxChoices !=1
            if(maxChoices != 1){
    	        if(cardinality.equals("single"))
    	    		throw new SpecificationViolationException("Cannot assign multiple choices to a single cardinality variable (" + responseIdentifier + ")");
    	    }
            
        	// get any previous values, i.e. user response, and store them in the vector
        	response = (Vector)interactionResponse.get("candidateValues");
        	candVals.add(i, response);
    		
    		// process the interaction
        	ints.add(i, processInteraction(interaction));
    	}
    	
    	// generate applet code
    	engine.addToOutput("<applet");
    	//TODO local URL
    	engine.addToOutput("codebase='" + basePath + "'", false);
    	engine.addToOutput("code = 'rhotspotV2.class'", false);
    	engine.addToOutput("name = 'TestAppletV2'", false);
    	engine.addToOutput("id = 'TestAppletV2'", false);
    	engine.addToOutput("width = '" + obj.getWidth() + "'", false);
    	engine.addToOutput("height = '" + (obj.getHeight() + 40) + "'", false);
    	engine.addToOutput("hspace = '0'", false);
    	engine.addToOutput("vspace = '0'", false);
    	engine.addToOutput("align = 'middle'", false);
    	engine.addToOutput(">", false);
    	
    	// add the parameters for the main image and overall interaction
    	engine.addToOutput("<param name = 'NoOfMainImages' value = '1'/>", false);
    	
    	String imagePath = obj.getData();
    	if(imagePath.startsWith(basePath))
    		imagePath = imagePath.substring(basePath.length());
    	engine.addToOutput("<param name ='background_image' value ='" + imagePath + "'>", false);   	
    	
    	engine.addToOutput("<param name = 'Mainimageno1' value = '" + imagePath + "::0::0::" + obj.getWidth() + "::" + obj.getHeight() + "'/>", false);
    	engine.addToOutput("<param name = 'baseType' value = 'point '/>", false);
    	engine.addToOutput("<param name = 'noOfTargets' value = '0'/>", false);
    	engine.addToOutput("<param name = 'identifiedTargets' value = 'FALSE'/>", false);
    	// engine.addToOutput("<param name = 'markerName' value = 'images/marker.gif'/>", false);
    	
    	// create the string of interaction responseIdentifiers
    	String interString = "";
    	
    	Vector interVec = (Vector)ints.elementAt(0);
    	String responseID = (String)interVec.elementAt(0);
    	interString = responseID;
    	
    	for(int i=1; i<ints.size(); i++) {
    		interVec = (Vector)ints.elementAt(i);
    		responseID = (String)interVec.elementAt(0);
    		interString = interString + "::" + responseID;    		
    	}
    	
    	engine.addToOutput("<param name = 'interactions' value = '" + interString + "'/>", false);
    	
    	// process each interaction and its associated image
    	for(int j=0; j<ints.size(); j++) {
    		
    		interVec = (Vector)ints.elementAt(j);
    		String respID = (String)interVec.elementAt(0);
    		String matches = (String)interVec.elementAt(1);
    		XHTMLObject imgObj = (XHTMLObject)interVec.elementAt(2);
    		
    		// add maxMatches tag for this interaction
    		engine.addToOutput("<param name = 'maxChoices:" + respID + "' value = '" + matches + "' />", false);
    		
    		imagePath = imgObj.getData();
        	if(imagePath.startsWith(basePath))
        		imagePath = imagePath.substring(basePath.length());
        	
        	// add marker tags for each possible match in the current interaction
    		for(int k=1; k<=Integer.parseInt(matches); k++) {
    			engine.addToOutput("<param name = 'labelNo" + k + ":" + respID + "' value = '::"
    					+ imgObj.getType() + "::" + imagePath + "::" + imgObj.getWidth() +
    					"::" + imgObj.getHeight() + "::" + matches + "' />", false);
    		}
    		
    		// add previous input, if any
        	String temp = "<param name = 'feedbackState:" + respID + "' value = '";
        	
        	if(candVals.elementAt(j) != null){
        		temp += "Yes";
        		Vector thisResponse = (Vector)candVals.elementAt(j);
        		
        		for(int l = 0; l<thisResponse.size(); l++) {
        			String resp = (String)thisResponse.elementAt(l);
        			temp += ":" + resp;
        		}
        	}
        	
        	else temp += "No";
        	temp += "'/>";
        	engine.addToOutput(temp, false);   		
    		
    	}
    	    	
    	engine.addToOutput("<param name = 'noOfMarkers' value ='" + maxChoices + "'/>", false);
    	engine.addToOutput("<param name = 'markerType' value ='LABELS'/>", false);
    	   	
    	//close the applet
    	engine.decreaseDepth();
    	engine.addToOutput("</applet>", false);
    	
	}
	
	private void processPrompt(Element _e) throws UnsupportedTagException, RenderingException, SpecificationViolationException {
		
		// add <div> tags and get the processingEngine to process the prompt
		engine.addToOutput("<div>");
		engine.processElements(_e.getContent());
		engine.decreaseDepth();
		engine.addToOutput("</div>", false);
		
	}
	
	protected XHTMLObject processObject(Element _e) throws RenderingException, SpecificationViolationException {
		
		// check that "data" and "type" attributes are both present
		String data = _e.getAttributeValue("data");
		if(data == null) 
			throw new SpecificationViolationException("'data' attribute not set in object");
		String type = _e.getAttributeValue("type");
		if(type == null) 
			throw new SpecificationViolationException("'type' attribute not set in object");
		
		// process and return the object
		ObjectProcessor op = new ObjectProcessor(supportedObjectTypes);		
		return op.processObject(_e);
		
	}
	
	private Vector processInteraction(Element _e) throws RenderingException, SpecificationViolationException {
			
		// find the object node for this interaction's image
    	Element object = _e.getChild("object", XHTMLRenderer.ns);
    	
    	if(object == null)
    		throw new SpecificationViolationException("No object found under SelectPointInteraction");
    	
    	// get this interaction's image (processed)
    	XHTMLObject obj = processObject(object);    	
		
		// get this interaction's details
		String data = _e.getAttributeValue("responseIdentifier");
		String choices = _e.getAttributeValue("maxChoices");
		
		
		// bundle processed interaction objects into an array
		Vector returnObjs = new Vector();
		returnObjs.add(0, data);
		returnObjs.add(1, choices);
		returnObjs.add(2, obj);
		
		// return the array
		return returnObjs;
		
	}
	
}
