/*
<LICENCE>

Copyright (c) 2007, University of Southampton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this
	list of conditions and the following disclaimer.

  *	Redistributions in binary form must reproduce the above copyright notice,
	this list of conditions and the following disclaimer in the documentation
	and/or other materials provided with the distribution.

  *	Neither the name of the University of Southampton nor the names of its
	contributors may be used to endorse or promote products derived from this
	software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

</LICENCE>
*/

package r2q2.rendering.xhtml.processors;

import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;

import org.jdom.Attribute;
import org.jdom.Element;

import com.gavinwillingham.util.properties.NoSuchPropertyException;
import com.gavinwillingham.util.properties.PropertiesLoader;

import r2q2.SpecificationViolationException;
import r2q2.rendering.RenderingException;
import r2q2.rendering.xhtml.UnsupportedTagException;
import r2q2.rendering.xhtml.XHTMLRenderer;

public class SliderInteractionProcessor extends InteractionProcessor {
	
	private double lowerBound = 0;
	private double upperBound = 0;
	private int step = 1;
	private boolean stepLabel = false;
	private String orientation = "Scrollbar.HORIZONTAL";
	private boolean reverse = false;
	private String previousAnswer = null;
	
	public SliderInteractionProcessor(XHTMLRenderer re) throws RenderingException {
		super(re);
		supportedTags.add("sliderInteraction");
	}
	
	public void process(Element _e) throws UnsupportedTagException, RenderingException, SpecificationViolationException{
		
		String elementName = _e.getName();
		
		if(!supportedTags.contains(elementName))
			throw new UnsupportedTagException(elementName + " in sliderInteraction");
		
		if(elementName.equals("sliderInteraction")) 
			processSliderInteraction(_e);
     }
	
	public void processSliderInteraction(Element _e) throws UnsupportedTagException, RenderingException, SpecificationViolationException{
        
		//process the prompt, if there is one
		int childSize = _e.getChildren().size(); 
		if(childSize > 1)
			throw new SpecificationViolationException("Too many children of sliderinteraction");
		if(childSize == 1){
			Element prompt = _e.getChild("prompt", engine.ns);
			if(prompt == null)
				throw new SpecificationViolationException("Child of sliderinteraction not prompt");
			engine.processElement(prompt);
		}
		
		//Now process all of the attributes in the array 
        Iterator<Attribute> attrs = _e.getAttributes().iterator();
        Attribute attr;                  
        String attname = "";                  
        String value; 
        try{
	        while (attrs.hasNext()){          
	            attr = attrs.next();                  
	            attname = attr.getName();                  
	            value = attr.getValue();  
	            if(attname.equals("responseIdentifier"))
	                responseIdentifier = value;
	            else if(attname.equals("lowerBound")){
	            	lowerBound = Double.parseDouble(value);
	            }
				else if(attname.equals("upperBound")){
					upperBound = Double.parseDouble(value);
				}
				else if(attname.equals("step")){
					step = Integer.parseInt(value);
				}
				else if(attname.equals("stepLabel")){
					stepLabel = Boolean.parseBoolean(value);
				}
				else if(attname.equals("orientation")){
					if(value.equals("horizontal"))
						orientation = "Scrollbar.HORIZONTAL";
					else if(value.equals("vertical"))
						orientation = "Scrollbar.VERTICAL";
					else 
						throw new SpecificationViolationException("Invalid orientation for slider (" + value + ")");
				}
				else if(attname.equals("reverse")){
					reverse = Boolean.parseBoolean(value);
				}
	        }
        }
        catch(NumberFormatException e){
        	throw new SpecificationViolationException(attname + " could not be converted to the appropriate type", e);
        }
        
        HashMap responseVar = (HashMap) engine.vars[0].get(responseIdentifier);
        String baseType = (String) responseVar.get("baseType");
        if(!baseType.equals("float") && !baseType.equals("integer"))
        	throw new SpecificationViolationException("basetype for a slider interaction must be float or integer, is " + baseType);
        String cardinality = (String) responseVar.get("cardinality");
        if(!cardinality.equals("single"))
        	throw new SpecificationViolationException("cardinality for a slider interaction must be single, is " + cardinality);
        
        String defaultValue = (String) responseVar.get("defaultValue");
        Vector candidateValues = (Vector) responseVar.get("candidateValues");
        if(candidateValues != null) previousAnswer = (String) candidateValues.get(0);
        
        //start the applet tag
        engine.addToOutput("<applet"); 
        try {
			engine.addToOutput("codebase='" + PropertiesLoader.getProperty("r2q2.router.content.httpBasePath") + "'", false);
		} catch (Exception e) {
			throw new RenderingException("Error loading htmlBasePath property", e);
		}
        engine.addToOutput("code='rendersliderV2.class'", false);
        engine.addToOutput("width = '600'", false); 
        engine.addToOutput("height = '100'", false);
        engine.addToOutput("hspace = '0'", false);
        engine.addToOutput("vspace = '0'", false);
        engine.addToOutput("align = 'middle'", false);
        engine.addToOutput(" >", false);
        
        engine.addToOutput("<param name = 'identifier' value = '" + responseIdentifier + "'/>", false);
        engine.addToOutput("<param name = 'baseType' value = '" + baseType + "'/>", false);
        engine.addToOutput("<param name = 'lowerBound' value = '" + lowerBound + "'/>", false);
        engine.addToOutput("<param name = 'upperBound' value = '" + upperBound + "'/>", false);
        engine.addToOutput("<param name = 'step' value = '" + step + "'>", false);
        engine.addToOutput("<param name = 'stepLabel' value = '" + stepLabel + "'/>", false);
        if(defaultValue != null)
        	engine.addToOutput("<param name = 'startVal' value = '" + defaultValue + "'/>", false);
        engine.addToOutput("<param name = 'orientation' value = '" + orientation + "'/>", false);
        
        //TODO what is this, it breaks without this here
        engine.addToOutput("<param name = 'NoOfChoices' value = '1'/>", false);
        
        //if we have a user response, set the initial value to it
        String temp = "<param name = 'feedbackState' value = '";
        if(previousAnswer != null) temp += "Yes:" + previousAnswer;
        else temp += "No";
        temp += "'/>";
        engine.addToOutput(temp, false);
        
        //close the applet tag
        engine.decreaseDepth();
        engine.addToOutput("</applet>", false);
	}

}
