/*
<LICENCE>

Copyright (c) 2007, University of Southampton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this
	list of conditions and the following disclaimer.

  *	Redistributions in binary form must reproduce the above copyright notice,
	this list of conditions and the following disclaimer in the documentation
	and/or other materials provided with the distribution.

  *	Neither the name of the University of Southampton nor the names of its
	contributors may be used to endorse or promote products derived from this
	software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

</LICENCE>
*/

package r2q2.router;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.Iterator;

import javax.activation.DataHandler;
import javax.xml.soap.AttachmentPart;
import javax.xml.soap.SOAPException;

import org.apache.axis.Message;
import org.apache.axis.MessageContext;
import org.apache.axis.client.Service;

import r2q2.NoAttachmentException;
import r2q2.SpecificationViolationException;
import r2q2.content.ws.ContentUnpackerServiceLocator;
import r2q2.content.ws.ContentUnpackerSoapBindingStub;
import r2q2.content.ws.EncodedFileList;
import r2q2.initialisation.InitialisationException;
import r2q2.processing.ProcessingException;
import r2q2.rendering.RenderingException;
import r2q2.router.interactions.InteractionProcessorController;
import r2q2.router.interactions.NoPreviousStageException;
import r2q2.router.interactions.NoSuchProcessorException;
import r2q2.util.soap.attachments.AttachmentBuilder;
import r2q2.util.streams.StreamCopier;

import com.gavinwillingham.util.properties.NoSuchPropertyException;
import com.gavinwillingham.util.properties.PropertiesLoader;

/**
 * TODO javadoc
 * @author Gavin Willingham
 */
public class Router {
    	
	private static String localBasePath = null;
    private static String httpBasePath = null;
    private static String contentUnpackerWebServiceAddress = null;

    private static InteractionProcessorController ipc = new InteractionProcessorController();
    
    /**
     * Default constructor. Loads properties if not already set
     * 
     * @throws IOException 	if the PropertiesLoader can't find the properties 
     * 						file
     * @throws NoSuchPropertyException 	if either the VM argument is not set, 
     * 									or one of the requested properties is 
     * 									not found
     */
    public Router() throws IOException, NoSuchPropertyException {
    	//If necessary, assign property values
        if(localBasePath == null) 
        	localBasePath = PropertiesLoader.getProperty("r2q2.router.content.localBasePath");
        if(httpBasePath == null) 
        	httpBasePath = PropertiesLoader.getProperty("r2q2.router.content.httpBasePath");
        if(contentUnpackerWebServiceAddress == null) 
        	contentUnpackerWebServiceAddress = PropertiesLoader.getProperty("r2q2.content.ws.location");
    }
    
    /**
     * Creates a new session.
     * 
     * <p>
     * Creates a new InteractionProcessor built upon the QTI XML content
     * package or XML file supplied as an attachment.
     * </p>
     * 
     * @param _xmlName 	This parameter is the filename of the question XML
     * @param _isContentPackage Specifies whether the attachment is a content 
     * 							package or xml file
     * @param _format The rendering format
     * 
     * @return GUID the unique identifier for the new session
     * 
     * TODO document exceptions
     * TODO don't throw so many exceptions
     * 
     * @throws IOException 
     * @throws SOAPException 
     * @throws MalformedURLException 
     * @throws NoSuchProcessorException 
     * @throws NoAttachmentException 
     * @throws InitialisationException 
     * @throws NoSuchPropertyException 
     * @throws IOException 
     * @throws SOAPException 
     * @throws MalformedURLException 
     * @throws RoutingException 
     */
    public String newSession(String _xmlName, boolean _isContentPackage, String _format) 
    	throws NoSuchProcessorException, NoAttachmentException, InitialisationException, NoSuchPropertyException, MalformedURLException, SOAPException, IOException, RoutingException {
    	
    	try{
	    	//Test for attachments
	        Message request = MessageContext.getCurrentContext().getRequestMessage();
	        if(request.countAttachments() == 0){			//there are noattachments
	        	throw new NoAttachmentException("No supplied content package");
	        }
	        Iterator attachmentIter = request.getAttachments();
	        AttachmentPart ap = (AttachmentPart)attachmentIter.next();
	    	
	    	if(_isContentPackage){
	        	String guid = ipc.newProcessor(_format);
	        	String encodedXmlName = unpack(ap, guid, _xmlName);
	        	byte[] xml = StreamCopier.copyToByteArray(new FileInputStream(new File(encodedXmlName)));
	        	ipc.initialiseProcessor(guid, xml);
	        	return guid;
	    	}
	    	else{
	    		String guid = ipc.newProcessor(_format);
	    		//TODO call content packer with byte[] not String
	    		byte[] xml = StreamCopier.copyToByteArray(ap.getDataHandler().getInputStream());
	    		ipc.initialiseProcessor(guid, xml);
	        	return guid;
	    	}
    	}
    	catch (NullPointerException e){
    		e.printStackTrace();
    		throw new RoutingException("Error creating new session", e);
    	}
    }    

    /**
     * TODO javadoc
     * 
     * @param _guid
     * @return 
     * @throws NoSuchProcessorException
     * @throws MalformedURLException
     * @throws IOException
     * @throws NoSuchPropertyException
     * @throws NoAttachmentException
     * @throws SOAPException
     * @throws SpecificationViolationException 
     * @throws RenderingException 
     */
    public HashMap firstStage(String _guid) throws NoSuchProcessorException, MalformedURLException, IOException, NoSuchPropertyException, NoAttachmentException, SOAPException, RenderingException, SpecificationViolationException{
    	try{
	    	RenderingResponse resp = ipc.firstStage(_guid);
	    	
	    	//Create the SOAP attachment
	        MessageContext msgContext = MessageContext.getCurrentContext();
	        Message msgResponse = msgContext.getResponseMessage();
	        AttachmentBuilder.addByteAttachment(msgResponse, resp.file, "question.xml");
	        
	        return resp.outcomeVars;
    	}
    	catch(NullPointerException e){
    		e.printStackTrace();
    		return null;
    	}
    }   
    
    /**
     * Calls the InteractionProcessor to process any response and render the
     * next stage for (potential) further interaction
     * 
     * TODO should probably be attaching the byte array here, not returning
     * 
     * @param _guid GUID of the InteractionProcessor
     * @return byte array containing the rendered output
     * @throws NoSuchProcessorException if an invalid GUID is supplied
     * @throws SOAPException 
     * @throws NoAttachmentException 
     * @throws NoSuchPropertyException 
     * @throws IOException 
     * @throws MalformedURLException 
     * @throws ProcessingException 
     * @throws InvalidUserResponseException 
     * @throws SpecificationViolationException 
     * @throws RenderingException 
     */
    public HashMap nextStage(String _guid, HashMap _repsonse) throws NoSuchProcessorException, MalformedURLException, IOException, NoSuchPropertyException, NoAttachmentException, SOAPException, RenderingException, SpecificationViolationException, InvalidUserResponseException, ProcessingException{
    	//TODO needs a test here on the HashMap, throw an InvalidArgumentException
    	
    	//call the IP
    	RenderingResponse resp = ipc.nextStage(_guid, _repsonse);
        
    	//Create the SOAP attachment
        MessageContext msgContext = MessageContext.getCurrentContext();
        Message msgResponse = msgContext.getResponseMessage();
        AttachmentBuilder.addByteAttachment(msgResponse, resp.file, "question.xml");
        
        return resp.outcomeVars;
    }
    
    /**
     * TODO javadoc
     * 
     * @param _guid
     * @throws NoSuchProcessorException
     * @throws MalformedURLException
     * @throws IOException
     * @throws NoSuchPropertyException
     * @throws NoAttachmentException
     * @throws SOAPException
     * @throws NoPreviousStageException 
     * @throws SpecificationViolationException 
     * @throws RenderingException 
     */
    public HashMap previousStage(String _guid) throws NoSuchProcessorException, MalformedURLException, IOException, NoSuchPropertyException, NoAttachmentException, SOAPException, NoPreviousStageException, RenderingException, SpecificationViolationException{
    	
    	RenderingResponse resp = ipc.previousStage(_guid);
    	
    	//Create the SOAP attachment
        MessageContext msgContext = MessageContext.getCurrentContext();
        Message msgResponse = msgContext.getResponseMessage();
        AttachmentBuilder.addByteAttachment(msgResponse, resp.file, "question.xml");
        
        return resp.outcomeVars;
    }
    
    public void terminateSession(String _guid) throws NoSuchProcessorException{
    	
    }
    
    /**
     * Calls the unpacking service
     * 
     * <p>
     * TODO: better description
     * </p>
     * 
     * @param _ap The AttachmentPart holding the content package
     * @param _guid The GUID of this user interaction
     * @param _xmlName 
     * @return The path to the newly-unpacked XML file
     * TODO document exceptions
     * @throws NoAttachmentException 
     * @throws RoutingException 
     */
    private String unpack(AttachmentPart _ap, String _guid, String _xmlName) throws SOAPException, 
                                                                  MalformedURLException, 
                                                                  IOException, NoAttachmentException, RoutingException {
        DataHandler dh = _ap.getDataHandler();
                
        //Initialise SOAP objects
        Service loc = new ContentUnpackerServiceLocator();
        ContentUnpackerSoapBindingStub stub = new ContentUnpackerSoapBindingStub(
                                                new URL(contentUnpackerWebServiceAddress),
                                                loc);
                                                
        //Add the attachment and invoke the method
        stub.addAttachment(dh);
        EncodedFileList filelist = stub.unpack(_xmlName, httpBasePath + _guid + "/");
        if(filelist == null)
    		throw new RoutingException("Content unpacker response empty!");
        Object[] attachments = stub.getAttachments();
        
        return processUnpackerResponse(filelist, attachments, _guid);
    }
    
    /**
     * TODO javadoc
     * 
     * @param _guid
     * @param _xml
     * @param _xmlName
     * @return
     * @throws IOException
     * @throws SOAPException
     * @throws NoAttachmentException
     * @throws RoutingException 
     */
    private String unpack(String _guid, String _xml, String _xmlName) 
    	throws IOException, SOAPException, NoAttachmentException, RoutingException{
    
    	//Initialise SOAP objects
        Service loc = new ContentUnpackerServiceLocator();
        ContentUnpackerSoapBindingStub stub = new ContentUnpackerSoapBindingStub(
                                                new URL(contentUnpackerWebServiceAddress),
                                                loc);
    	
        //invoke method
        EncodedFileList filelist = stub.collateFiles(_xml, _xmlName, httpBasePath);
        if(filelist == null)
    		throw new RoutingException("Content unpacker response empty!");
        Object[] attachments = stub.getAttachments();
    	
    	return processUnpackerResponse(filelist, attachments, _guid);
    }
    
    /**
     * TODO javadoc
     * 
     * @param _filelist
     * @param _attachments
     * @return
     * @throws IOException
     * @throws SOAPException
     * @throws NoAttachmentException
     * @throws RoutingException 
     */
    private String processUnpackerResponse(EncodedFileList _filelist, Object[] _attachments, String _guid) 
    	throws IOException, SOAPException, NoAttachmentException, RoutingException{
        
    	if(_filelist == null)
    		throw new RoutingException("Content unpacker response empty!");
    	
    	String[] filenames = _filelist.getFilenames();
        
        //Throw exception if nothing returned
        if(filenames == null 
        	|| filenames.length == 0)
        	//TODO throw an exception
        	return null;
        	;
        
        /* Throw exception if nothing attached. If this line is reached, the 
         * content unpacker is broken, because it shouldn't return file names
         * but no files.
         */
        if(_attachments == null
        	|| _attachments.length == 0)
        	throw new NoAttachmentException("No attachments returned from content unpacker");
        
        //Create the base directory
        String path = localBasePath + _guid + File.separator;
        File dir = new File(path);
        dir.mkdirs();
        
        //Store the files
        for(int i=0, j=0; i<filenames.length; i++){
        	
        	File file = new File(path + filenames[i]);
        	        	
	        //create the file	
            if (filenames[i].endsWith("\\") || filenames[i].endsWith("/")) { //If a directory, create it
                file.mkdir();
            } 
            else {  //If a file, write it out
            	try{
	                AttachmentPart responseAttachment = (AttachmentPart)_attachments[j++];
	                InputStream is = responseAttachment.getDataHandler().getInputStream();
	                OutputStream os = new FileOutputStream(file);
	                StreamCopier.copy(is,os);
	                is.close();
	                os.close();
            	}
            	catch(FileNotFoundException e){
            		e.printStackTrace();
            	}
            }
        }
        
        return path + _filelist.getXmlName();
    }
}
