/*
<LICENCE>

Copyright (c) 2007, University of Southampton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this
	list of conditions and the following disclaimer.

  *	Redistributions in binary form must reproduce the above copyright notice,
	this list of conditions and the following disclaimer in the documentation
	and/or other materials provided with the distribution.

  *	Neither the name of the University of Southampton nor the names of its
	contributors may be used to endorse or promote products derived from this
	software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

</LICENCE>
*/

package r2q2.router.interactions;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Vector;

import javax.activation.DataHandler;
import javax.xml.soap.SOAPException;

import org.apache.axis.AxisFault;

import com.gavinwillingham.util.properties.NoSuchPropertyException;
import com.gavinwillingham.util.properties.PropertiesLoader;

import r2q2.NoAttachmentException;
import r2q2.SpecificationViolationException;
import r2q2.initialisation.InitialisationException;
import r2q2.initialisation.ws.InitialisationEngineServiceLocator;
import r2q2.initialisation.ws.InitialisationEngineSoapBindingStub;
import r2q2.processing.ProcessingException;
import r2q2.processing.ws.ProcessingEngineServiceLocator;
import r2q2.processing.ws.ProcessingEngineSoapBindingStub;
import r2q2.rendering.RenderingEngineResponse;
import r2q2.rendering.RenderingException;
import r2q2.rendering.ws.RenderingEngineClient;
import r2q2.router.InvalidUserResponseException;
import r2q2.router.RenderingResponse;
import r2q2.util.streams.ByteArrayDataSource;

/**
 * The main processing class within the Router. One InteractionProcessor is
 * created per user-item interaction. This deals with calling the individual
 * web services, and stores the state of the interaction.
 * 
 * @author Gavin Willingham
 */
public class InteractionProcessor {
    
    private String guid = null;
    private byte[] xml = null;
    private String format;
    
    private LinkedList<StageVariables> variables = new LinkedList<StageVariables>();
    
    private boolean initialised = false;
    
    /**
     * Constructs a new InteractionProcessor with only a GUID. If instantiating
     * using this method, the initialise(String, String, String) method MUST be
     * called prior to any processing method
     * TODO: or a ProcessorNotInitialisedException will be thrown
     * 
     * @param _guid The unique identifier for this object
     * @param _format The rendering format
     */
    public InteractionProcessor(String _guid, String _format) {    	
    	
        this.guid = _guid;
        this.format = _format;
    }    
    
    /**
     * Constructs a new InteractionProcessor. Variables are initialised at this 
     * stage.
     * 
     * @param _guid The unique identifier for this object
     * @param _format The rendering format
     * @param _xml The XML definition of this question
     * 
     * @throws InitialisationException 
     * @throws NoSuchPropertyException 
     * @throws IOException 
     * @throws MalformedURLException 
     */
    public InteractionProcessor(String _guid, String _format, byte[] _xml) throws InitialisationException, MalformedURLException, IOException, NoSuchPropertyException {    	
    	this(_guid, _format);
        initialise(_xml);
    }

    /**
     * Stores the XML, filename and base path then calls the initialisation
     * engine to initialise the variables, and the processing engine to 
     * perfom any necessary template processing
     * 
     * TODO javadoc exceptions
     * 
     * @param _xml The XML definition of this question
     * @throws InitialisationException 
     * @throws NoSuchPropertyException 
     * @throws IOException 
     * @throws MalformedURLException 
     */
    public void initialise(byte[] _xml) throws InitialisationException, MalformedURLException, IOException, NoSuchPropertyException{
    	if(!initialised){
    		//TODO if any NULL, throw an exception, remove =null from declarations
    		this.xml = _xml;
    		
	        //Call the initialisation engine
    		InitialisationEngineServiceLocator loc = new InitialisationEngineServiceLocator();
        	URL url = new URL(PropertiesLoader.getProperty("r2q2.initialisation.ws.location"));
        	InitialisationEngineSoapBindingStub stub = new InitialisationEngineSoapBindingStub(url, loc);
	        HashMap[] temp = stub.intialiseQuestion(new String(_xml));
	        	        
	        //process the response
	        //TODO: it should be the initialisation engine throwing this
	        if(temp == null) throw new InitialisationException("Variable array is null!");
	        	        
	        //Call the processing engine for template processing
	        ProcessingEngineServiceLocator loc2 = new ProcessingEngineServiceLocator();
	    	URL url2 = new URL(PropertiesLoader.getProperty("r2q2.processing.ws.location"));
	    	ProcessingEngineSoapBindingStub stub2 = new ProcessingEngineSoapBindingStub(url2, loc2);
	    	stub2.addAttachment(
	    			new DataHandler(
	    				new ByteArrayDataSource("question.xml", xml, "text/xml")));
	        temp = stub2.processTemplate(temp);
	    	
	    	//Store the initialised variables
	        StageVariables sv = new StageVariables(temp, null, null);
	    	variables.add(sv);
	    	
	    	initialised = true;
    	}
    }
    
    /**
     * TODO javadoc
     * 
     * @return
     * @throws MalformedURLException
     * @throws IOException
     * @throws NoSuchPropertyException
     * @throws NoAttachmentException 
     * @throws SOAPException 
     * @throws SpecificationViolationException 
     * @throws RenderingException 
     */
    public synchronized RenderingResponse firstStage() throws RenderingException, SpecificationViolationException{
    	if(variables.size() > 1){
    		StageVariables sv = variables.getFirst();
    		variables = new LinkedList<StageVariables>();
    		variables.add(sv);
    	}
    	return callRenderer(true);
    }
    
    /**
     * Calls the processing engine to update the variable values, then invokes
     * the rendering engine to render the next stage
     * 
     * TODO throw exception if not initialised
     * TODO javadoc exceptions
     * 
     * @param _response Vector containing the user's response
     * @return byte array containing the rendered output
     * @throws NoSuchPropertyException 
     * @throws IOException 
     * @throws MalformedURLException 
     * @throws SOAPException 
     * @throws NoAttachmentException 
     * @throws SpecificationViolationException 
     * @throws RenderingException 
     * @throws InvalidUserResponseException 
     * @throws ProcessingException 
     */
    public synchronized RenderingResponse nextStage(HashMap _response) throws RenderingException, SpecificationViolationException, InvalidUserResponseException, ProcessingException{
    	//Can call this method and not use firstStage if desired
    	if(variables.size() < 2) return firstStage();
    	
    	//Clone variables
    	HashMap[] oldVars = variables.get(variables.size() - 2).vars;
    	HashMap[] newVars = new HashMap[3];
    	newVars[0] = (HashMap)oldVars[0].clone();
    	newVars[1] = (HashMap)oldVars[1].clone();
    	newVars[2] = (HashMap)oldVars[2].clone();
    	
    	//Update the variables
    	HashMap<String, HashMap> responseVars = (HashMap<String, HashMap>)newVars[0];
    	Iterator<String> iter = _response.keySet().iterator();
    	while(iter.hasNext()){
    		String nextKey = iter.next();
    		//HashMap rDV = responseVars.get(nextKey);
    		HashMap rDV = (HashMap) responseVars.get(nextKey).clone();
    		if(rDV == null){
    			System.out.println("Unknown response variable: " + nextKey);
    			throw new InvalidUserResponseException("Unknown response variable: " + nextKey);
    		}
    		else{
    			//TODO validation of user response
    			Object temp = _response.get(nextKey);
    			if(temp instanceof Vector)
    				rDV.put("candidateValues", temp);
    			else if (temp instanceof String){
    				Vector v = new Vector();
    				v.add(temp);
    				rDV.put("candidateValues", v);
    			}
    			else {
    				throw new InvalidUserResponseException("Unknown response format for variable: " + nextKey);
    			}
    		}
    		responseVars.put(nextKey, rDV);
    	}
    	
    	//Call the processing engine for response processing
    	//TODO extract out to client class
        ProcessingEngineServiceLocator loc2 = new ProcessingEngineServiceLocator();
    	URL url2 = null;
		try {
			url2 = new URL(PropertiesLoader.getProperty("r2q2.processing.ws.location"));
		} catch (MalformedURLException e) {
			throw new ProcessingException(e);
		} catch (IOException e) {
			throw new ProcessingException(e);
		} catch (NoSuchPropertyException e) {
			throw new ProcessingException(e);
		}
    	ProcessingEngineSoapBindingStub stub2;
		try {
			stub2 = new ProcessingEngineSoapBindingStub(url2, loc2);
		} catch (AxisFault e) {
			throw new ProcessingException(e);
		}
    	stub2.addAttachment(
    			new DataHandler(
    				new ByteArrayDataSource("question.xml", xml, "text/xml")));
        try {
			newVars = stub2.processResponse(newVars);
		} catch (r2q2.processing.ws.NoAttachmentException e) {
			throw new ProcessingException(e);
		} catch (r2q2.processing.ws.ProcessingException e) {
			throw new ProcessingException(e);
		} catch (r2q2.processing.ws.SpecificationViolationException e) {
			throw new SpecificationViolationException(e);
		} catch (RemoteException e) {
			throw new ProcessingException(e);
		}
    	
        //Store the updated variables
        StageVariables sv = variables.removeLast();
        sv.vars = newVars;
    	variables.add(sv);
    	
    	//Render the next stage
    	return callRenderer(false);
    }
    
    public synchronized RenderingResponse previousStage() throws NoPreviousStageException, RenderingException, SpecificationViolationException{
    	if(variables.size() < 3) throw new NoPreviousStageException();
    	//delete last 2 entries in list
    	variables.removeLast();
    	variables.removeLast();
    	
    	boolean flag = false;
    	if(variables.size() == 1) flag = true;
    	
    	//call rendering engine
    	return callRenderer(flag);
    }
    
    private RenderingResponse callRenderer(boolean _firstStage) throws RenderingException, SpecificationViolationException{
    	//get the most recent variables
    	StageVariables oldsv = variables.getLast();
    	System.out.println("Stage " + variables.size());
    	System.out.println("vars = " + oldsv.vars);
    	System.out.println("opts = " + oldsv.options);
    	System.out.println("varnames = " + oldsv.varnames);
    	
    	//Call the rendering service
    	RenderingResponse out_ = new RenderingResponse(null, oldsv.vars[1]);
    	RenderingEngineClient reClient;
		try {
			reClient = new RenderingEngineClient(PropertiesLoader.getProperty("r2q2.rendering.ws.location"));
		} catch (IOException e) {
			throw new RenderingException("Couldn't read properties file");
		} catch (NoSuchPropertyException e) {
			throw new RenderingException("r2q2.rendering.ws.location property not set");
		}
    	RenderingEngineResponse resp = reClient.render(xml, oldsv.vars, oldsv.options, oldsv.varnames, _firstStage, format, out_);
    	   	
    	//Process the response
    	//store the options
    	StageVariables newsv = new StageVariables(null,  resp.getOrder(), resp.getVarNames());
    	variables.add(newsv);
    	
    	return out_;
    }
}

//TODO perhaps not the most elegant way to do it
class StageVariables{
	HashMap[] vars;
	HashMap options;
	List<String> varnames;
	
	StageVariables(HashMap[] _vars, HashMap _opts, List<String> _varnames){ vars = _vars; options = _opts; varnames = _varnames;}
}
