/*
<LICENCE>

Copyright (c) 2007, University of Southampton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this
	list of conditions and the following disclaimer.

  *	Redistributions in binary form must reproduce the above copyright notice,
	this list of conditions and the following disclaimer in the documentation
	and/or other materials provided with the distribution.

  *	Neither the name of the University of Southampton nor the names of its
	contributors may be used to endorse or promote products derived from this
	software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

</LICENCE>
*/

package r2q2.router.interactions;

import java.io.IOException;
import java.net.MalformedURLException;
import java.util.HashMap;
import java.util.UUID;
import java.util.Vector;

import javax.xml.soap.SOAPException;

import com.gavinwillingham.util.properties.NoSuchPropertyException;

import r2q2.NoAttachmentException;
import r2q2.SpecificationViolationException;
import r2q2.initialisation.InitialisationException;
import r2q2.processing.ProcessingException;
import r2q2.rendering.RenderingException;
import r2q2.router.InvalidUserResponseException;
import r2q2.router.RenderingResponse;

/**
 * Provides abstracted access to the InteractionProcessors. Provides methods 
 * for creating and accessing the processors, addressable by GUID.
 * 
 * @see r2q2.router.interactions.InteractionProcessor
 * @author Gavin Willingham
 */
public class InteractionProcessorController {
    
	private static HashMap<String,InteractionProcessor> processors = new HashMap<String,InteractionProcessor>();

    /**
     * Creates a new processor when XML is not available. Note that if created
     * by this method, a processor must be initialised by calling the 
     * initialiseProcessor(String,String,String,String) method.
     * 
     * @return GUID of new InteractionProcessor
     */
    public String newProcessor(String _format){
        String guid_ = makeGUID();
        InteractionProcessor ip = new InteractionProcessor(guid_, _format);
        processors.put(guid_, ip);
        return guid_;
    }
    
    /**
     * Creates a new InteractionProcessor for the supplied question
     * 
     * @param _xml The question XML file
     * @param _filename The filename of the XML file
     * @return GUID of new InteractionProcessor
     * @throws InitialisationException 
     * @throws NoSuchPropertyException 
     * @throws IOException 
     * @throws MalformedURLException 
     */
    public String newProcessor(byte[] _xml, String _format) throws InitialisationException, MalformedURLException, IOException, NoSuchPropertyException{
        String guid_ = makeGUID();
        InteractionProcessor ip = new InteractionProcessor(guid_, _format, _xml);
        processors.put(guid_, ip);
        return guid_;
    }
    
    /**
     * Initialises the relevant processor
     * 
     * @param _guid The GUID of the processor to initialise
     * @param _xml The XML definition of this question
     * @param _filename The name of the Item's XML file
     * @throws NoSuchProcessorException if an invalid GUID is supplied
     * @throws InitialisationException 
     * @throws NoSuchPropertyException 
     * @throws IOException 
     * @throws MalformedURLException 
     */
    public void initialiseProcessor(String _guid, byte[] _xml) throws NoSuchProcessorException, InitialisationException, MalformedURLException, IOException, NoSuchPropertyException{
    	InteractionProcessor ip = processors.get(_guid);
    	if(ip == null) throw new NoSuchProcessorException("No processor for GUID: " + _guid);
    	ip.initialise(_xml);
    }
    
    /**
     * TODO javadoc
     * 
     * @param _guid
     * @return
     * @throws NoSuchProcessorException
     * @throws MalformedURLException
     * @throws IOException
     * @throws NoSuchPropertyException
     * @throws NoAttachmentException
     * @throws SOAPException
     * @throws SpecificationViolationException 
     * @throws RenderingException 
     */
    public RenderingResponse firstStage(String _guid) throws NoSuchProcessorException, NoAttachmentException, RenderingException, SpecificationViolationException{
    	InteractionProcessor ip = processors.get(_guid);
    	if(ip == null) throw new NoSuchProcessorException("No processor for GUID: " + _guid);
    	return ip.firstStage();
    }
    
    /**
     * Calls the nextStage() method of the appropriate processor
     * 
     * TODO javadoc
     * 
     * @param _guid the GUID of the processor to call
     * @return the rendered output of the next stage
     * @throws NoSuchProcessorException if an invalid GUID is supplied
     * @throws SpecificationViolationException 
     * @throws RenderingException 
     * @throws ProcessingException 
     * @throws InvalidUserResponseException 
     * @throws SOAPException 
     * @throws NoAttachmentException 
     * @throws NoSuchPropertyException 
     * @throws IOException 
     * @throws MalformedURLException 
     */
    public RenderingResponse nextStage(String _guid, HashMap _response) throws NoSuchProcessorException, RenderingException, SpecificationViolationException, InvalidUserResponseException, ProcessingException{
    	InteractionProcessor ip = processors.get(_guid);
    	if(ip == null) throw new NoSuchProcessorException("No processor for GUID: " + _guid);
    	return ip.nextStage(_response);
    }
    
    /**
     * TODO javadoc
     * 
     * @param _guid
     * @return
     * @throws NoSuchProcessorException
     * @throws MalformedURLException
     * @throws IOException
     * @throws NoSuchPropertyException
     * @throws NoAttachmentException
     * @throws SOAPException
     * @throws NoPreviousStageException
     * @throws SpecificationViolationException 
     * @throws RenderingException 
     */
    public RenderingResponse previousStage(String _guid) throws NoSuchProcessorException, MalformedURLException, IOException, NoSuchPropertyException, NoAttachmentException, SOAPException, NoPreviousStageException, RenderingException, SpecificationViolationException{
    	InteractionProcessor ip = processors.get(_guid);
    	if(ip == null) throw new NoSuchProcessorException("No processor for GUID: " + _guid);
    	return ip.previousStage();
    }
    
    /**
     * Generates a unique identfier
     * 
     * <p>
     * Uses java.util.UUID and the &quot;Int-&quot; prefix
     * <p>
     * 
     * @return Globally unique identifier
     */
    private synchronized String makeGUID(){
        String guid_;
        while(true){    //loop on the (num_generated in 2^122) chance that the generated GUID is a duplicate
            guid_ = "Int-" + UUID.randomUUID().toString();
            if(!processors.containsKey(guid_)) return guid_;
        }
    }
}
